@@ -5,43 +5,36 @@ ways.
 
 import ast
 import base64
-import logging # isort:skip
+import logging
 import os
+import pathlib
 
 from glob import glob
 from types import ModuleType
 
+from bokeh.application import Application
+from bokeh.application.handlers.document_lifecycle import (
+    DocumentLifecycleHandler,
+)
+from bokeh.application.handlers.function import FunctionHandler
 from bokeh.command.subcommands.serve import Serve as _BkServe
 from bokeh.command.util import build_single_handler_applications
-
-from bokeh.application import Application
-from bokeh.application.handlers.document_lifecycle import DocumentLifecycleHandler
-from bokeh.application.handlers.function import FunctionHandler
 from bokeh.server.contexts import ApplicationContext
-from tornado.ioloop import PeriodicCallback, IOLoop
+from tornado.ioloop import PeriodicCallback
 from tornado.web import StaticFileHandler
 
 from ..auth import OAuthProvider
 from ..config import config
-from ..util import fullpath
-from ..io.rest import REST_PROVIDERS
+from ..io.document import _cleanup_doc
+from ..io.liveness import LivenessHandler
 from ..io.reload import record_modules, watch
+from ..io.rest import REST_PROVIDERS
 from ..io.server import INDEX_HTML, get_static_routes, set_curdoc
 from ..io.state import state
+from ..util import fullpath
 
 log = logging.getLogger(__name__)
 
-
-def _cleanup_doc(doc):
-    for callback in doc.session_destroyed_callbacks:
-        try:
-            callback(None)
-        except Exception:
-            pass
-    doc.callbacks._change_callbacks[None] = {}
-    doc.destroy(None)
-
-
 def parse_var(s):
     """
     Parse a key, value pair, separated by '='
@@ -70,10 +63,9 @@ def parse_vars(items):
 class AdminApplicationContext(ApplicationContext):
 
     def __init__(self, application, unused_timeout=15000, **kwargs):
-        super().__init__(application, io_loop=IOLoop.current(), **kwargs)
+        super().__init__(application, **kwargs)
         self._unused_timeout = unused_timeout
         self._cleanup_cb = None
-        self._loop.add_callback(self.run_load_hook)
 
     async def cleanup_sessions(self):
         await self._cleanup_sessions(self._unused_timeout)
@@ -144,6 +136,11 @@ class Serve(_BkServe):
             help    = "Expiry off the OAuth cookie in number of days.",
             default = 1
         )),
+        ('--auth-template', dict(
+            action  = 'store',
+            type    = str,
+            help    = "Template to serve when user is unauthenticated."
+        )),
         ('--rest-provider', dict(
             action = 'store',
             type   = str,
@@ -176,7 +173,7 @@ class Serve(_BkServe):
         ('--profiler', dict(
             action  = 'store',
             type    = str,
-            help    = "The profiler to use by default, e.g. pyinstrument or snakeviz."
+            help    = "The profiler to use by default, e.g. pyinstrument, snakeviz or memray."
         )),
         ('--autoreload', dict(
             action  = 'store_true',
@@ -193,7 +190,17 @@ class Serve(_BkServe):
             type    = str,
             help    = "Path to a setup script to run before server starts.",
             default = None
-        ))
+        )),
+        ('--liveness', dict(
+            action  = 'store_true',
+            help    = "Whether to add a liveness endpoint."
+        )),
+        ('--liveness-endpoint', dict(
+            action  = 'store',
+            type    = str,
+            help    = "The endpoint for the liveness API.",
+            default = "liveness"
+        )),
     )
 
     # Supported file extensions
@@ -294,12 +301,10 @@ class Serve(_BkServe):
                         state._on_load(None)
                     _cleanup_doc(doc)
 
-        prefix = args.prefix
-        if prefix is None:
-            prefix = ""
-        prefix = prefix.strip("/")
-        if prefix:
-            prefix = "/" + prefix
+        if args.liveness:
+            argvs = {f: args.args for f in files}
+            applications = build_single_handler_applications(files, argvs)
+            patterns += [(r"/%s" % args.liveness_endpoint, LivenessHandler, dict(applications=applications))]
 
         config.profiler = args.profiler
         if args.admin:
@@ -308,14 +313,15 @@ class Serve(_BkServe):
             config._admin = True
             app = Application(FunctionHandler(admin_panel))
             unused_timeout = args.check_unused_sessions or 15000
-            app_ctx = AdminApplicationContext(app, unused_timeout=unused_timeout, url='/admin')
+            state._admin_context = app_ctx = AdminApplicationContext(
+                app, unused_timeout=unused_timeout, url='/admin'
+            )
             if all(not isinstance(handler, DocumentLifecycleHandler) for handler in app._handlers):
                 app.add(DocumentLifecycleHandler())
             app_patterns = []
             for p in per_app_patterns:
                 route = '/admin' + p[0]
                 context = {"application_context": app_ctx}
-                route = prefix + route
                 app_patterns.append((route, p[1], context))
 
             websocket_path = None
@@ -350,8 +356,22 @@ class Serve(_BkServe):
                 )
             config.nthreads = args.num_threads
 
+        if args.auth_template:
+            authpath = pathlib.Path(args.auth_template)
+            if not authpath.isfile():
+                raise ValueError(
+                    "The supplied auth-template {args.auth_template} does not "
+                    "exist, ensure you supply and existing Jinja2 template."
+                )
+            config.auth_template = str(authpath.absolute())
+        if args.oauth_provider and config.oauth_provider:
+                raise ValueError(
+                    "Supply OAuth provider either using environment variable "
+                    "or via explicit argument, not both."
+                )
         if args.oauth_provider:
             config.oauth_provider = args.oauth_provider
+        if config.oauth_provider:
             config.oauth_expiry = args.oauth_expiry_days
             if config.oauth_key and args.oauth_key:
                 raise ValueError(
@@ -411,7 +431,7 @@ class Serve(_BkServe):
                       "encryption key and will potentially leak "
                       "credentials in cookies and a JWT token embedded "
                       "in the served website. Use at your own risk or "
-                      "generate a key with the `panel oauth-key` CLI "
+                      "generate a key with the `panel oauth-secret` CLI "
                       "command and then provide it to `panel serve` "
                       "using the PANEL_OAUTH_ENCRYPTION environment "
                       "variable or the --oauth-encryption-key CLI "
@@ -442,7 +462,13 @@ class Serve(_BkServe):
                     "CLI argument or the PANEL_COOKIE_SECRET environment "
                     "variable."
                 )
-            kwargs['auth_provider'] = OAuthProvider(error_template=args.oauth_error_template)
+            if args.oauth_error_template:
+                error_template = str(pathlib.Path(args.oauth_error_template).absolute())
+            elif config.auth_template:
+                error_template = config.auth_template
+            else:
+                error_template = None
+            kwargs['auth_provider'] = OAuthProvider(error_template=error_template)
 
             if args.oauth_redirect_uri and config.oauth_redirect_uri:
                 raise ValueError(
