@@ -1,6 +1,7 @@
 """
 Utilities for building custom models included in panel.
 """
+import fnmatch
 import glob
 import inspect
 import io
@@ -8,6 +9,7 @@ import os
 import pathlib
 import shutil
 import tarfile
+import zipfile
 
 import param
 import requests
@@ -76,6 +78,8 @@ def require_components():
     return configs, requirements, exports, skip_import
 
 def write_bundled_files(name, files, bundle_dir, explicit_dir=None, ext=None):
+    from .config import config
+
     model_name = name.split('.')[-1].lower()
     for bundle_file in files:
         bundle_file = bundle_file.split('?')[0]
@@ -96,15 +100,23 @@ def write_bundled_files(name, files, bundle_dir, explicit_dir=None, ext=None):
                 map_response = requests.get(map_file, verify=False)
             except Exception:
                 map_response = None
-        bundle_path = os.path.join(*os.path.join(*bundle_file.split('//')[1:]).split('/')[1:])
+        if bundle_file.startswith(config.npm_cdn):
+            bundle_path = os.path.join(*bundle_file.replace(config.npm_cdn, '').split('/'))
+        else:
+            bundle_path = os.path.join(*os.path.join(*bundle_file.split('//')[1:]).split('/')[1:])
         obj_dir = explicit_dir or model_name
         filename = bundle_dir.joinpath(obj_dir, bundle_path)
         filename.parent.mkdir(parents=True, exist_ok=True)
         filename = str(filename)
         if ext and not str(filename).endswith(ext):
             filename += f'.{ext}'
-        with open(filename, 'w', encoding="utf-8") as f:
-            f.write(response.content.decode('utf-8'))
+        if filename.endswith('.ttf'):
+            with open(filename, 'wb') as f:
+                f.write(response.content)
+        else:
+            content = response.content.decode('utf-8')
+            with open(filename, 'w', encoding="utf-8") as f:
+                f.write(content)
         if map_response:
             with open(f'{filename}.map', 'w', encoding="utf-8") as f:
                 f.write(map_response.content.decode('utf-8'))
@@ -124,26 +136,58 @@ def write_bundled_tarball(name, tarball, bundle_dir, module=False):
         if not tarf.name.startswith(tarball['src']) or not tarf.isfile():
             continue
         path = tarf.name.replace(tarball['src'], '')
-        if any(path.startswith(exc) for exc in exclude):
+        if any(fnmatch.fnmatch(tarf.name, exc) for exc in exclude):
             continue
         bundle_path = os.path.join(*path.split('/'))
-        dest_path = os.path.join(*tarball['dest'].split('/'))
+        dest_path = tarball['dest'].replace('/', os.path.sep)
         filename = bundle_dir.joinpath(model_name, dest_path, bundle_path)
         filename.parent.mkdir(parents=True, exist_ok=True)
         fobj = tar_obj.extractfile(tarf.name)
-        content = fobj.read().decode('utf-8')
         filename = str(filename)
         if module and filename.endswith('.js'):
             filename = filename[:-3]
             if filename.endswith('index'):
                 filename += '.mjs'
-        with open(filename, 'w', encoding="utf-8") as f:
-            f.write(content)
+        if any(filename.endswith(ft) for ft in ('.ttf', '.eot', '.woff', '.woff2')):
+            content = fobj.read()
+            with open(filename, 'wb') as f:
+                f.write(content)
+        else:
+            content = fobj.read().decode('utf-8')
+            with open(filename, 'w', encoding="utf-8") as f:
+                f.write(content)
     tar_obj.close()
 
+def write_bundled_zip(name, resource, bundle_dir):
+    try:
+        response = requests.get(resource['zip'])
+    except Exception:
+        response = requests.get(resource['zip'], verify=False)
+    f = io.BytesIO()
+    f.write(response.content)
+    f.seek(0)
+    zip_obj = zipfile.ZipFile(f)
+    exclude = resource.get('exclude', [])
+    for zipf in zip_obj.namelist():
+        path = zipf.replace(resource['src'], '')
+        if any(fnmatch.fnmatch(zipf, exc) for exc in exclude) or zipf.endswith('/'):
+            continue
+        bundle_path = path.replace('/', os.path.sep)
+        filename = bundle_dir.joinpath(name, bundle_path)
+        filename.parent.mkdir(parents=True, exist_ok=True)
+        fdata = zip_obj.read(zipf)
+        filename = str(filename)
+        if any(filename.endswith(ft) for ft in ('.ttf', '.eot', '.woff', '.woff2')):
+            with open(filename, 'wb') as f:
+                f.write(fdata)
+        else:
+            with open(filename, 'w', encoding="utf-8") as f:
+                f.write(fdata.decode('utf-8'))
+    zip_obj.close()
 
-def bundle_resources():
+def bundle_resources(verbose=False):
     from .config import panel_extension
+    from .io.resources import RESOURCE_URLS
     from .reactive import ReactiveHTML
     from .template.base import BasicTemplate
     from .template.theme import Theme
@@ -154,6 +198,15 @@ def bundle_resources():
 
     bundle_dir = pathlib.Path(__file__).parent.joinpath('dist', 'bundled')
 
+    # Collect shared resources
+    for name, resource in RESOURCE_URLS.items():
+        if verbose:
+            print(f'Bundling shared resource {name}.')
+        if 'zip' in resource:
+            write_bundled_zip(name, resource, bundle_dir)
+        elif 'tar' in resource:
+            write_bundled_tarball(name, resource, bundle_dir)
+
     # Extract Model dependencies
     js_files, css_files, resource_files = {}, {}, {}
     reactive = param.concrete_descendents(ReactiveHTML).values()
@@ -164,6 +217,8 @@ def bundle_resources():
     for name, model in models:
         if not name.startswith('panel.'):
             continue
+        if verbose:
+            print(f'Collecting {name} resources')
         prev_jsfiles = getattr(model, '__javascript_raw__', None)
         prev_jsbundle = getattr(model, '__tarball__', None)
         prev_cls = model
@@ -198,12 +253,16 @@ def bundle_resources():
 
     # Bundle Model dependencies
     for name, jsfiles in js_files.items():
+        if verbose:
+            print(f'Bundling {name} model JS resources')
         if isinstance(jsfiles, dict):
             write_bundled_tarball(name, jsfiles, bundle_dir)
         else:
             write_bundled_files(name, jsfiles, bundle_dir)
 
     for name, cssfiles in css_files.items():
+        if verbose:
+            print(f'Bundling {name} model CSS resources')
         write_bundled_files(name, cssfiles, bundle_dir)
 
     for name, res_files in resource_files.items():
@@ -211,6 +270,8 @@ def bundle_resources():
 
     # Bundle Template resources
     for name, template in param.concrete_descendents(BasicTemplate).items():
+        if verbose:
+            print(f'Bundling {name} resources')
         # Bundle Template._resources
         if template._resources.get('bundle', True):
             write_bundled_files(name, list(template._resources.get('css', {}).values()), bundle_dir, 'css')
@@ -221,7 +282,7 @@ def bundle_resources():
                     js_modules.append(js_module)
             write_bundled_files(name, js_modules, bundle_dir, 'js', ext='mjs')
             for tarball in template._resources.get('tarball', {}).values():
-                write_bundled_tarball('js', tarball, bundle_dir, module=True)
+                write_bundled_tarball('js', tarball, bundle_dir)
 
         # Bundle CSS files in template dir
         template_dir = pathlib.Path(inspect.getfile(template)).parent
@@ -266,6 +327,8 @@ def bundle_resources():
 
     # Bundle Theme classes
     for name, theme in param.concrete_descendents(Theme).items():
+        if verbose:
+            print(f'Bundling {name} theme resources')
         if theme.base_css:
             theme_bundle_dir = bundle_dir / theme.param.base_css.owner.__name__.lower()
             theme_bundle_dir.mkdir(parents=True, exist_ok=True)
@@ -274,3 +337,10 @@ def bundle_resources():
             tmplt_bundle_dir = bundle_dir / theme._template.__name__.lower()
             tmplt_bundle_dir.mkdir(parents=True, exist_ok=True)
             shutil.copyfile(theme.css, tmplt_bundle_dir / os.path.basename(theme.css))
+
+    # Bundle icons & images
+    dest_dir = bundle_dir / 'images'
+    dest_dir.mkdir(parents=True, exist_ok=True)
+    icon_dir = pathlib.Path(__file__).parent.parent / 'doc' / '_static' / 'icons'
+    for icon in glob.glob(str(icon_dir / '*')):
+        shutil.copyfile(icon, dest_dir / os.path.basename(icon))
