@@ -16,11 +16,10 @@ from weakref import WeakKeyDictionary
 import param
 
 from pyviz_comms import (
-    JupyterCommManager as _JupyterCommManager, extension as _pyviz_extension
+    JupyterCommManager as _JupyterCommManager, extension as _pyviz_extension,
 )
 
 from .io.logging import panel_log_handler
-from .io.notebook import load_notebook
 from .io.state import state
 
 __version__ = str(param.version.Version(
@@ -91,14 +90,38 @@ class _config(_base_config):
     >>> pn.config.loading_spinner = 'bar'
     """
 
+    admin_plugins = param.List([], item_type=tuple, doc="""
+        A list of tuples containing a title and a function that returns
+        an additional panel to be rendered into the admin page.""")
+
     apply_signatures = param.Boolean(default=True, doc="""
         Whether to set custom Signature which allows tab-completion
         in some IDEs and environments.""")
 
+    authorize_callback = param.Callable(default=None, doc="""
+        Authorization callback that is invoked when authentication
+        is enabled. The callback is given the user information returned
+        by the configured Auth provider and should return True or False
+        depending on whether the user is authorized to access the
+        application.""")
+
+    auth_template = param.Path(default=None, doc="""
+        A jinja2 template rendered when the authorize_callback determines
+        that a user in not authorized to access the application.""")
+
     autoreload = param.Boolean(default=False, doc="""
         Whether to autoreload server when script changes.""")
 
-    loading_spinner = param.Selector(default='arcs', objects=[
+    defer_load = param.Boolean(default=False, doc="""
+        Whether to defer load of rendered functions.""")
+
+    exception_handler = param.Callable(default=None, doc="""
+        General exception handler for events.""")
+
+    load_entry_points = param.Boolean(default=True, doc="""
+        Load entry points from external packages.""")
+
+    loading_spinner = param.Selector(default='arc', objects=[
         'arc', 'arcs', 'bar', 'dots', 'petal'], doc="""
         Loading indicator to use when component loading parameter is set.""")
 
@@ -112,7 +135,7 @@ class _config(_base_config):
         Whether to enable notifications functionality.""")
 
     profiler = param.Selector(default=None, allow_None=True, objects=[
-        'pyinstrument', 'snakeviz'], doc="""
+        'pyinstrument', 'snakeviz', 'memray'], doc="""
         The profiler engine to enable.""")
 
     safe_embed = param.Boolean(default=False, doc="""
@@ -175,6 +198,12 @@ class _config(_base_config):
         default='WARNING', objects=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
         doc="Log level of Panel loggers")
 
+    _npm_cdn = param.Selector(default='https://cdn.jsdelivr.net/npm',
+        objects=['https://unpkg.com', 'https://cdn.jsdelivr.net/npm'],  doc="""
+        The CDN to load NPM packages from if resources are served from
+        CDN. Allows switching between https://unpkg.com and
+        https://cdn.jsdelivr.net/npm for most resources.""")
+
     _nthreads = param.Integer(default=None, doc="""
         When set to a non-None value a thread pool will be started.
         Whenever an event arrives from the frontend it will be
@@ -209,11 +238,12 @@ class _config(_base_config):
         Whether to inline JS and CSS resources. If disabled, resources
         are loaded from CDN if one is available.""")
 
+    # Global parameters that are shared across all sessions
     _globals = [
-        'autoreload', 'comms', 'cookie_secret', 'nthreads',
-        'oauth_provider', 'oauth_expiry', 'oauth_key', 'oauth_secret',
-        'oauth_jwt_user', 'oauth_redirect_uri','oauth_encryption_key',
-        'oauth_extra_params'
+        'admin_plugins', 'autoreload', 'comms', 'cookie_secret',
+        'nthreads', 'oauth_provider', 'oauth_expiry', 'oauth_key',
+        'oauth_secret', 'oauth_jwt_user', 'oauth_redirect_uri',
+        'oauth_encryption_key', 'oauth_extra_params', 'npm_cdn'
     ]
 
     _truthy = ['True', 'true', '1', True, 1]
@@ -246,6 +276,9 @@ class _config(_base_config):
     def _enable_notifications(self):
         from .io.notifications import NotificationArea
         from .io.state import state
+        from .reactive import ReactiveHTMLMetaclass
+        if self.notifications and 'notifications' not in ReactiveHTMLMetaclass._loaded_extensions:
+            ReactiveHTMLMetaclass._loaded_extensions.add('notifications')
         if not state.curdoc:
             state._notification = NotificationArea()
 
@@ -293,6 +326,12 @@ class _config(_base_config):
         panel_log_handler.setLevel(self._log_level)
 
     def __getattribute__(self, attr):
+        """
+        Ensures that configuration parameters that are defined per
+        session are stored in a per-session dictionary. This is to
+        ensure that even on first access mutable parameters do not
+        end up being modified.
+        """
         from .io.state import state
         init = super().__getattribute__('initialized')
         global_params = super().__getattribute__('_globals')
@@ -301,16 +340,17 @@ class _config(_base_config):
         else:
             params = []
         session_config = super().__getattribute__('_session_config')
-        if state.curdoc and state.curdoc not in session_config:
-            session_config[state.curdoc] = {}
+        curdoc = state.curdoc
+        if curdoc and curdoc not in session_config:
+            session_config[curdoc] = {}
         if (attr in ('raw_css', 'css_files', 'js_files', 'js_modules') and
-            state.curdoc and attr not in session_config[state.curdoc]):
+            curdoc and attr not in session_config[curdoc]):
             new_obj = copy.copy(super().__getattribute__(attr))
             setattr(self, attr, new_obj)
         if attr in global_params:
             return super().__getattribute__(attr)
-        elif state.curdoc and state.curdoc in session_config and attr in session_config[state.curdoc]:
-            return session_config[state.curdoc][attr]
+        elif curdoc and curdoc in session_config and attr in session_config[curdoc]:
+            return session_config[curdoc][attr]
         elif f'_{attr}' in params and getattr(self, f'_{attr}_') is not None:
             return super().__getattribute__(f'_{attr}_')
         return super().__getattribute__(attr)
@@ -367,6 +407,10 @@ class _config(_base_config):
         log_level = os.environ.get('PANEL_LOG_LEVEL', self._log_level)
         return log_level.upper() if log_level else None
 
+    @property
+    def npm_cdn(self):
+        return os.environ.get('PANEL_NPM_CDN', _config._npm_cdn)
+
     @property
     def nthreads(self):
         nthreads = os.environ.get('PANEL_NUM_THREADS', self._nthreads)
@@ -469,7 +513,7 @@ class panel_extension(_pyviz_extension):
         'terminal': 'panel.models.terminal',
         'tabulator': 'panel.models.tabulator',
         'texteditor': 'panel.models.quill',
-        'jsoneditor': 'panel.models.json_editor'
+        'jsoneditor': 'panel.models.jsoneditor'
     }
 
     # Check whether these are loaded before rendering (if any item
@@ -537,7 +581,7 @@ class panel_extension(_pyviz_extension):
 
         if 'holoviews' in sys.modules:
             import holoviews as hv
-            import holoviews.plotting.bokeh # noqa
+            import holoviews.plotting.bokeh  # noqa
             loaded = loaded or getattr(hv.extension, '_loaded', False)
 
             if hv.Store.current_backend in hv.Store.renderers:
@@ -555,6 +599,8 @@ class panel_extension(_pyviz_extension):
         except Exception:
             return
 
+        from .io.notebook import load_notebook
+
         newly_loaded = [arg for arg in args if arg not in panel_extension._loaded_extensions]
         if loaded and newly_loaded:
             self.param.warning(
@@ -596,7 +642,7 @@ class panel_extension(_pyviz_extension):
 
         # Try to detect environment so that we can enable comms
         try:
-            import google.colab # noqa
+            import google.colab  # noqa
             config.comms = "colab"
             return
         except ImportError:
@@ -612,13 +658,19 @@ class panel_extension(_pyviz_extension):
         if config.notifications:
             display(state.notifications) # noqa
 
+        if config.load_entry_points:
+            self._load_entry_points()
+
     def _apply_signatures(self):
         from inspect import Parameter, Signature
+
         from .viewable import Viewable
 
         descendants = param.concrete_descendents(Viewable)
         for cls in reversed(list(descendants.values())):
-            if cls.__doc__.startswith('params'):
+            if cls.__doc__ is None:
+                pass
+            elif cls.__doc__.startswith('params'):
                 prefix = cls.__doc__.split('\n')[0]
                 cls.__doc__ = cls.__doc__.replace(prefix, '')
             sig = inspect.signature(cls.__init__)
@@ -648,6 +700,15 @@ class panel_extension(_pyviz_extension):
                 parameters, return_annotation=sig.return_annotation
             )
 
+    def _load_entry_points(self):
+        ''' Load entry points from external packages
+        Import is performed here so any importlib/pkg_resources
+        can be easily bypassed by switching off the configuration flag.
+        also, no reason to waste time on importing this module
+        if it wont be used.
+        '''
+        from .entry_points import load_entry_points
+        load_entry_points('panel.extension')
 
 #---------------------------------------------------------------------
 # Private API
