@@ -1,21 +1,26 @@
 from __future__ import annotations
 
-import datetime as dt
+import asyncio
 import dataclasses
+import datetime as dt
 import inspect
+import logging
 import threading
 
 from contextlib import contextmanager
 from functools import partial, wraps
-from typing import Callable, Iterator, List, Optional
+from typing import (
+    Callable, Iterator, List, Optional,
+)
 
 from bokeh.application.application import SessionContext
 from bokeh.document.document import Document
 from bokeh.document.events import DocumentChangedEvent, ModelChangedEvent
-from bokeh.io import curdoc as _curdoc
 
 from .model import monkeypatch_events
-from .state import set_curdoc, state
+from .state import curdoc_locked, state
+
+logger = logging.getLogger(__name__)
 
 #---------------------------------------------------------------------
 # Private API
@@ -54,12 +59,49 @@ def _dispatch_events(doc: Document, events: List[DocumentChangedEvent]) -> None:
     for event in events:
         doc.callbacks.trigger_on_change(event)
 
+def _cleanup_doc(doc):
+    for callback in doc.session_destroyed_callbacks:
+        try:
+            callback(None)
+        except Exception:
+            pass
+    doc.callbacks._change_callbacks[None] = {}
+
+    # Remove views
+    from ..viewable import Viewable
+    views = {}
+    for ref, (pane, root, vdoc, comm) in list(state._views.items()):
+        if vdoc is doc:
+            pane._cleanup(root)
+            if isinstance(pane, Viewable):
+                pane._hooks = []
+                for p in pane.select():
+                    p._hooks = []
+                    p._param_watchers = {}
+                    p._documents = {}
+                    p._callbacks = {}
+            pane._param_watchers = {}
+            pane._documents = {}
+            pane._callbacks = {}
+        else:
+            views[ref] = (pane, root, doc, comm)
+    state._views = views
+
+    # Clean up templates
+    if doc in state._templates:
+        tmpl = state._templates[doc]
+        tmpl._documents = {}
+        del state._templates[doc]
+
+    # Destroy doc
+    doc.destroy(None)
+
 #---------------------------------------------------------------------
 # Public API
 #---------------------------------------------------------------------
 
 def init_doc(doc: Optional[Document]) -> Document:
-    curdoc = doc or _curdoc()
+    curdoc = doc or curdoc_locked()
     if not isinstance(curdoc, Document):
         curdoc = curdoc._doc
     if not curdoc.session_context:
@@ -67,8 +109,7 @@ def init_doc(doc: Optional[Document]) -> Document:
 
     thread = threading.current_thread()
     if thread:
-        with set_curdoc(curdoc):
-            state._thread_id = thread.ident
+        state._thread_id_[curdoc] = thread.ident
 
     session_id = curdoc.session_context.id
     sessions = state.session_info['sessions']
@@ -107,6 +148,40 @@ def with_lock(func: Callable) -> Callable:
     wrapper.lock = True # type: ignore
     return wrapper
 
+def dispatch_tornado(conn, event):
+    from tornado.websocket import WebSocketHandler
+    socket = conn._socket
+    ws_conn = socket.ws_connection
+    if not ws_conn or ws_conn.is_closing(): # type: ignore
+        return []
+    msg = conn.protocol.create('PATCH-DOC', [event])
+    futures = [
+        WebSocketHandler.write_message(socket, msg.header_json),
+        WebSocketHandler.write_message(socket, msg.metadata_json),
+        WebSocketHandler.write_message(socket, msg.content_json)
+    ]
+    for header, payload in msg._buffers:
+        futures.extend([
+            WebSocketHandler.write_message(socket, header),
+            WebSocketHandler.write_message(socket, payload, binary=True)
+        ])
+    return futures
+
+def dispatch_django(conn, event):
+    socket = conn._socket
+    msg = conn.protocol.create('PATCH-DOC', [event])
+    futures = [
+        socket.send(text_data=msg.header_json),
+        socket.send(text_data=msg.metadata_json),
+        socket.send(text_data=msg.content_json)
+    ]
+    for header, payload in msg._buffers:
+        futures.extend([
+            socket.send(text_data=header),
+            socket.send(binary_data=payload)
+        ])
+    return futures
+
 @contextmanager
 def unlocked() -> Iterator:
     """
@@ -120,7 +195,7 @@ def unlocked() -> Iterator:
     if (curdoc is None or session_context is None or session is None):
         yield
         return
-    from tornado.websocket import WebSocketHandler
+    from tornado.websocket import WebSocketClosedError, WebSocketHandler
     connections = session._subscribed_connections
 
     hold = curdoc.callbacks.hold_value
@@ -141,24 +216,29 @@ def unlocked() -> Iterator:
 
         events = curdoc.callbacks._held_events
         monkeypatch_events(events)
-        remaining_events = []
+        remaining_events, futures = [], []
         for event in events:
             if not isinstance(event, ModelChangedEvent) or event in old_events or locked:
                 remaining_events.append(event)
                 continue
             for conn in connections:
-                socket = conn._socket
-                ws_conn = getattr(socket, 'ws_connection', False)
-                if (not hasattr(socket, 'write_message') or
-                    ws_conn is None or (ws_conn and ws_conn.is_closing())): # type: ignore
-                    continue
-                msg = conn.protocol.create('PATCH-DOC', [event])
-                WebSocketHandler.write_message(socket, msg.header_json)
-                WebSocketHandler.write_message(socket, msg.metadata_json)
-                WebSocketHandler.write_message(socket, msg.content_json)
-                for header, payload in msg._buffers:
-                    WebSocketHandler.write_message(socket, header)
-                    WebSocketHandler.write_message(socket, payload, binary=True)
+                if isinstance(conn._socket, WebSocketHandler):
+                    futures += dispatch_tornado(conn, event)
+                else:
+                    futures += dispatch_django(conn, event)
+
+        # Ensure that all write_message calls are awaited and handled
+        async def handle_write_errors():
+            for future in futures:
+                try:
+                    await future
+                except WebSocketClosedError:
+                    logger.warning("Failed sending message as connection was closed")
+                except Exception as e:
+                    logger.warning(f"Failed sending message due to following error: {e}")
+
+        asyncio.ensure_future(handle_write_errors())
+
         curdoc.callbacks._held_events = remaining_events
     finally:
         if hold:
