@@ -7,19 +7,19 @@ import textwrap
 
 from contextlib import contextmanager
 from typing import (
-    TYPE_CHECKING, Any, List, Iterable, Optional
+    TYPE_CHECKING, Any, Iterable, List, Optional,
 )
-from typing_extensions import Literal
 
 import numpy as np
 
 from bokeh.document import Document
 from bokeh.document.events import (
-    ColumnDataChangedEvent, DocumentPatchedEvent, ModelChangedEvent
+    ColumnDataChangedEvent, DocumentPatchedEvent, ModelChangedEvent,
 )
 from bokeh.model import DataModel
 from bokeh.models import Box, ColumnDataSource, Model
 from bokeh.protocol import Protocol
+from typing_extensions import Literal
 
 from .state import state
 
@@ -77,6 +77,8 @@ def diff(
         return None
 
     patch_events = [event for event in events if isinstance(event, DocumentPatchedEvent)]
+    if not patch_events:
+        return
     monkeypatch_events(events)
     msg_type: Literal["PATCH-DOC"] = "PATCH-DOC"
     msg = Protocol().create(msg_type, patch_events, use_buffers=binary)
