@@ -1,17 +1,21 @@
+import io
 import os
 import re
+import tempfile
+import uuid
 
 from contextlib import contextmanager
 from cProfile import Profile
 from functools import wraps
 
+from ..config import config
 from ..util import escape
 from .state import state
 
 
 def render_pyinstrument(sessions, timeline=False, show_all=False):
-    from pyinstrument.session import Session
     from pyinstrument.renderers import HTMLRenderer
+    from pyinstrument.session import Session
     r = HTMLRenderer(timeline=timeline, show_all=show_all)
     session = sessions[0]
     if not timeline:
@@ -19,11 +23,11 @@ def render_pyinstrument(sessions, timeline=False, show_all=False):
             session = Session.combine(session, s)
     return escape(r.render(session)), ""
 
-
 def render_snakeviz(name, sessions):
+    from pstats import Stats
+
     import snakeviz
 
-    from pstats import Stats
     from snakeviz.stats import json_stats, table_rows
     from tornado.template import Template
 
@@ -38,6 +42,50 @@ def render_snakeviz(name, sessions):
     ).decode('utf-8').replace('/static/', '/snakeviz/static/')
     return escape(rendered), "background-color: white;"
 
+def render_memray(name, sessions, show_memory_leaks=True, merge_threads=True, reporter='tree'):
+    from memray import FileReader
+    from memray.reporters.flamegraph import FlameGraphReporter
+    from memray.reporters.stats import StatsReporter
+    from memray.reporters.table import TableReporter
+    from memray.reporters.tree import TreeReporter
+
+    reporter_cls = {
+        'flamegraph': FlameGraphReporter,
+        'stats': StatsReporter,
+        'table': TableReporter,
+        'tree': TreeReporter
+    }.get(reporter)
+
+    session = sessions[-1]
+    with tempfile.NamedTemporaryFile() as nf:
+        nf.write(session)
+        reader = FileReader(nf.name)
+        if show_memory_leaks:
+            snapshot = reader.get_leaked_allocation_records(
+                merge_threads=merge_threads if merge_threads is not None else True
+            )
+        else:
+            snapshot = reader.get_high_watermark_allocation_records(
+                merge_threads=merge_threads if merge_threads is not None else True
+            )
+
+        kwargs = {'native_traces': reader.metadata.has_native_traces}
+        if reporter in ('flamegraph', 'table'):
+            kwargs['memory_records'] = tuple(reader.get_memory_snapshots())
+        reporter_inst = reporter_cls.from_snapshot(
+            snapshot,
+            **kwargs
+        )
+
+    out = io.StringIO()
+    if reporter == 'flamegraph':
+        reporter_inst.render(out, reader.metadata, show_memory_leaks, merge_threads)
+    elif reporter == 'table':
+        reporter_inst.render(out, reader.metadata, show_memory_leaks)
+    else:
+        reporter_inst.render(file=out)
+    out.seek(0)
+    return out.read(), ""
 
 def get_profiles(profilers, **kwargs):
     from ..pane import HTML, Markdown
@@ -45,6 +93,16 @@ def get_profiles(profilers, **kwargs):
     for (path, engine), sessions in profilers.items():
         if not sessions:
             continue
+        if engine == 'memray':
+            src, style = render_memray(path, sessions, **kwargs)
+            if kwargs.get('reporter', 'tree') not in ('flamegraph', 'table'):
+                from ..widgets import Terminal
+                term = Terminal(sizing_mode='stretch_both', margin=0, min_height=600)
+                term.write(src)
+                profiles.append((path, term))
+                continue
+            else:
+                src = escape(src)
         if engine == 'pyinstrument':
             src, style = render_pyinstrument(sessions, **kwargs)
         elif engine == 'snakeviz':
@@ -53,7 +111,7 @@ def get_profiles(profilers, **kwargs):
             f'<iframe srcdoc="{src}" width="100%" height="100%" frameBorder="0" style="{style}"></iframe>',
             sizing_mode='stretch_both',
             margin=0,
-            min_height=400
+            min_height=800
         )
         profiles.append((path, html))
     if not profiles:
@@ -67,31 +125,45 @@ def get_sessions(allow=None, deny=None):
 
 
 def profiling_tabs(state, allow=None, deny=[]):
-    from ..layout import Accordion, Column, Row, Tabs
-    from ..widgets import Checkbox
+    from ..layout import (
+        Accordion, Column, Row, Tabs,
+    )
+    from ..widgets import Checkbox, Select
     tabs = Tabs(
         *get_profiles(get_sessions(allow, deny)),
         margin=(0, 5),
         sizing_mode='stretch_width'
     )
-    def update_profiles(*args):
+    def update_profiles(*args, **kwargs):
         tabs[:] = get_profiles(
-            get_sessions(allow, deny),
-            timeline=timeline.value,
-            show_all=show_all.value
+            get_sessions(allow, deny), **kwargs
+
         )
     state.param.watch(update_profiles, '_profiles')
-    timeline = Checkbox(name='Enable timeline', margin=(5, 0))
-    timeline.param.watch(update_profiles, 'value')
-    show_all = Checkbox(name='Show All', margin=(5, 0))
-    show_all.param.watch(update_profiles, 'value')
+
+    if config.profiler == 'pyinstrument':
+        def update_pyinstrument(*args):
+            update_profiles(timeline=timeline.value, show_all=show_all.value)
+        timeline = Checkbox(name='Enable timeline', margin=(5, 0))
+        timeline.param.watch(update_pyinstrument, 'value')
+        show_all = Checkbox(name='Show All', margin=(5, 0))
+        show_all.param.watch(update_pyinstrument, 'value')
+        config_panel = Row(
+            timeline,
+            show_all,
+            sizing_mode='stretch_width'
+        )
+    elif config.profiler == 'memray':
+        def update_memray(*args):
+            update_profiles(reporter=reporter.value)
+        reporter = Select(name='Reporter', options=['flamegraph', 'table', 'tree'], value='tree')
+        reporter.param.watch(update_memray, 'value')
+        config_panel = reporter
+    else:
+        config_panel = Row(sizing_mode='stretch_width')
     return Column(
         Accordion(
-            ('Config', Row(
-                timeline,
-                show_all,
-                sizing_mode='stretch_width'
-            )),
+            ('Config', config_panel),
             active=[],
             active_header_background='#444444',
             header_background='#333333',
@@ -113,7 +185,7 @@ def profile_ctx(engine='pyinstrument'):
     Arguments
     ---------
     engine: str
-      The profiling engine, e.g. 'pyinstrument' or 'snakeviz'
+      The profiling engine, e.g. 'pyinstrument', 'snakeviz' or 'memray'
 
     Returns
     -------
@@ -131,6 +203,11 @@ def profile_ctx(engine='pyinstrument'):
     elif engine == 'snakeviz':
         prof = Profile()
         prof.enable()
+    elif engine == 'memray':
+        import memray
+        tmp_file = f'{tempfile.gettempdir()}/tmp{uuid.uuid4().hex}'
+        tracker = memray.Tracker(tmp_file)
+        tracker.__enter__()
     elif engine is None:
         pass
     sessions = []
@@ -140,6 +217,10 @@ def profile_ctx(engine='pyinstrument'):
     elif engine == 'snakeviz':
         prof.disable()
         sessions.append(prof)
+    elif engine == 'memray':
+        tracker.__exit__(None, None, None)
+        sessions.append(open(tmp_file, 'rb').read())
+        os.remove(tmp_file)
 
 
 def profile(name, engine='pyinstrument'):
@@ -152,7 +233,7 @@ def profile(name, engine='pyinstrument'):
     name: str
       A unique name for the profiling session.
     engine: str
-      The profiling engine, e.g. 'pyinstrument' or 'snakeviz'
+      The profiling engine, e.g. 'pyinstrument', 'snakeviz' or 'memray'
     """
     if not isinstance(name, str):
         raise ValueError("Profiler must be given a name.")
