@@ -6,9 +6,9 @@ import types
 from contextlib import contextmanager
 from functools import partial
 
+from ..util import fullpath
 from .callbacks import PeriodicCallback
 from .state import state
-from ..util import fullpath
 
 _watched_files = set()
 _modules = set()
@@ -71,7 +71,7 @@ def autoreload_watcher():
     files and sys.modules.
     """
     cb = partial(_reload_on_update, {})
-    _callbacks[state.curdoc] = pcb = PeriodicCallback(callback=cb)
+    _callbacks[state.curdoc] = pcb = PeriodicCallback(callback=cb, background=True)
     pcb.start()
 
 def watch(filename):
