@@ -9,14 +9,14 @@ import warnings
 import weakref
 
 from typing import (
-    TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Tuple, Type, Union
+    TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Tuple, Type,
+    Union,
 )
 
-from bokeh.models import CustomJS, LayoutDOM
-from bokeh.models import Model as BkModel
-
 import param
 
+from bokeh.models import CustomJS, LayoutDOM, Model as BkModel
+
 from .io.datamodel import create_linked_datamodel
 from .models import ReactiveHTML
 from .reactive import Reactive
@@ -384,7 +384,7 @@ class CallbackGenerator(object):
                     model_spec = None
                 model = obj.handles[handle_spec]
         elif isinstance(obj, Viewable):
-            model, _ = obj._models[root_model.ref['id']]
+            model, _ = obj._models.get(root_model.ref['id'], (None, None))
         elif isinstance(obj, BkModel):
             model = obj
         elif isinstance(obj, param.Parameterized):
