@@ -6,10 +6,10 @@ import {ModelEvent, JSON} from "@bokehjs/core/bokeh_events"
 import {div} from "@bokehjs/core/dom"
 import {Enum} from "@bokehjs/core/kinds"
 import * as p from "@bokehjs/core/properties";
-import {ColumnDataSource} from "@bokehjs/models/sources/column_data_source";
+import {ColumnDataSource} from "@bokehjs/models/sources/column_data_source"
 import {TableColumn} from "@bokehjs/models/widgets/tables"
 
-import {debounce} from  "debounce"
+import {debounce} from "debounce"
 
 import {comm_settings} from "./comm_manager"
 import {transform_cds_to_records} from "./data"
@@ -120,6 +120,53 @@ function group_data(records: any[], columns: any[], indexes: string[], aggregato
   return grouped
 }
 
+
+const timestampSorter = function(a: any, b: any, _aRow: any, _bRow: any, _column: any, _dir: any, _params: any){
+  // Bokeh serializes datetime objects as UNIX timestamps.
+
+  //a, b - the two values being compared
+  //aRow, bRow - the row components for the values being compared (useful if you need to access additional fields in the row data for the sort)
+  //column - the column component for the column being sorted
+  //dir - the direction of the sort ("asc" or "desc")
+  //sorterParams - sorterParams object from column definition array
+
+  // Added an _ in front of some parameters as they're unused and the Typescript compiler was complaining about it.
+
+  // const alignEmptyValues = params.alignEmptyValues
+  let emptyAlign: any
+  emptyAlign = 0
+
+  const opts = {zone: new (window as any).luxon.IANAZone('UTC')}
+
+  // NaN values are serialized to -9223372036854776 by Bokeh
+
+  if (String(a) == '-9223372036854776') {
+    a = (window as any).luxon.DateTime.fromISO('invalid')
+  } else {
+    a = (window as any).luxon.DateTime.fromMillis(a, opts)
+  }
+  if (String(b) == '-9223372036854776') {
+    b = (window as any).luxon.DateTime.fromISO('invalid')
+  } else {
+    b = (window as any).luxon.DateTime.fromMillis(b, opts)
+  }
+
+  if(!a.isValid){
+    emptyAlign = !b.isValid ? 0 : -1;
+  }else if(!b.isValid){
+    emptyAlign =  1;
+  }else{
+    //compare valid values
+    return a - b;
+  }
+
+  // Invalid (e.g. NaN) always at the bottom
+  emptyAlign *= -1
+
+  return emptyAlign;
+}
+
+
 const dateEditor = function(cell: any, onRendered: any, success: any, cancel: any) {
   //cell - the cell component for the editable cell
   //onRendered - function to call when the editor has been rendered
@@ -238,43 +285,55 @@ export class DataTabulatorView extends PanelHTMLBoxView {
   _initializing: boolean
   _lastVerticalScrollbarTopPosition: number = 0;
   _applied_styles: boolean = false
+  _building: boolean=false
 
   connect_signals(): void {
     super.connect_signals()
 
-    const {configuration, layout, columns, theme, groupby} = this.model.properties;
+    const p = this.model.properties
+    const {configuration, layout, columns, theme, groupby} = p;
     this.on_change([configuration, layout, columns, groupby], debounce(() => this.invalidate_render(), 20, false))
 
     this.on_change([theme], () => this.setCSS())
 
-    this.connect(this.model.properties.download.change, () => {
+    this.connect(p.download.change, () => {
       const ftype = this.model.filename.endsWith('.json') ? "json" : "csv"
       this.tabulator.download(ftype, this.model.filename)
     })
 
-    this.connect(this.model.properties.children.change, () => this.renderChildren())
+    this.connect(p.children.change, () => this.renderChildren())
 
-    this.connect(this.model.properties.expanded.change, () => {
+    this.connect(p.expanded.change, () => {
+      // The first cell is the cell of the frozen _index column.
       for (const row of this.tabulator.rowManager.getRows()) {
         if (row.cells.length > 0)
           row.cells[0].layoutElement()
       }
+      // Make sure the expand icon is changed when expanded is
+      // changed from Python.
+      for (const row of this.tabulator.rowManager.getRows()) {
+        if (row.cells.length > 0) {
+          const index = row.data._index
+          const icon = this.model.expanded.indexOf(index) < 0 ? "►" : "▼"
+          row.cells[1].element.innerText = icon
+        }
+      }
     })
 
-    this.connect(this.model.properties.styles.change, () => {
+    this.connect(p.styles.change, () => {
       if (this._applied_styles)
-	this.tabulator.redraw(true)
+     this.tabulator.redraw(true)
       this.setStyles()
     })
-    this.connect(this.model.properties.hidden_columns.change, () => this.setHidden())
-    this.connect(this.model.properties.page_size.change, () => this.setPageSize())
-    this.connect(this.model.properties.page.change, () => {
+    this.connect(p.hidden_columns.change, () => this.setHidden())
+    this.connect(p.page_size.change, () => this.setPageSize())
+    this.connect(p.page.change, () => {
       if (!this._updating_page)
         this.setPage()
     })
-    this.connect(this.model.properties.max_page.change, () => this.setMaxPage())
-    this.connect(this.model.properties.frozen_rows.change, () => this.setFrozen())
-    this.connect(this.model.properties.sorters.change, () => this.setSorters())
+    this.connect(p.max_page.change, () => this.setMaxPage())
+    this.connect(p.frozen_rows.change, () => this.setFrozen())
+    this.connect(p.sorters.change, () => this.setSorters())
     this.connect(this.model.source.properties.data.change, () => this.setData())
     this.connect(this.model.source.streaming, () => this.addData())
     this.connect(this.model.source.patching, () => {
@@ -294,7 +353,7 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       sorters.push({column: '_index', dir: 'asc'})
     for (const sort of this.model.sorters.reverse()) {
       if (sort.column === undefined)
-	sort.column = sort.field
+        sort.column = sort.field
       sorters.push(sort)
     }
     return sorters
@@ -307,21 +366,16 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     this.relayout()
   }
 
-  renderComplete(): void {
-    // Only have to set up styles after initial render subsequent
-    // styling is handled by change event on styles property
-    if (this._initializing) {
-      this.setSelection()
-      this.relayout()
-      this._initializing = false
-      if (this._lastVerticalScrollbarTopPosition)
-	this.tabulator.rowManager.element.scrollTop = this._lastVerticalScrollbarTopPosition;
-    }
-  }
-
   redraw(): void {
-    this.tabulator.redraw(true)
-    this.setStyles()
+    if (!this._building) {
+      if (this.tabulator.columnManager.element != null) {
+	this.tabulator.columnManager.redraw(true);
+      }
+      if (this.tabulator.rowManager.renderer != null) {
+	this.tabulator.rowManager.redraw(true);
+	this.setStyles()
+      }
+    }
   }
 
   after_layout(): void {
@@ -341,50 +395,31 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     let configuration = this.getConfiguration()
 
     this.tabulator = new Tabulator(container, configuration)
-
+    this.init_callbacks()
     this.renderChildren(true)
 
-     // Swap pagination mode
-    if (this.model.pagination === 'remote') {
-      this.tabulator.options.pagination = this.model.pagination
-      this.tabulator.modules.page.mode = 'remote'
-    }
-
-    this.setGroupBy()
     this.setHidden()
 
-    // Set up page
-    if (this.model.pagination) {
-      this.setMaxPage()
-      this.tabulator.setPage(this.model.page)
-      this.setData()
-    } else
-      this.setFrozen()
-
     this.el.appendChild(container)
   }
 
-  /*
+  tableInit(): void {
+    this._building = true
+    // Patch the ajax request and page data parsing methods
+    const ajax = this.tabulator.modules.ajax
+    ajax.sendRequest = (_url: any, params: any, _config: any) => {
+      return this.requestPage(params.page, params.sort)
+    }
+    this.tabulator.modules.page._parseRemoteData = (): boolean => {
+      return false
+    }
+  }
+
   init_callbacks(): void {
     // Initialization
     this.tabulator.on("tableBuilding", () => this.tableInit())
     this.tabulator.on("tableBuilt", () => this.tableBuilt())
 
-    // For disabled pagination initialize on renderComplete
-    if (this.model.pagination !== 'remote') {
-      this.tabulator.on("renderComplete", () => {
-        this.tabulator.off("renderComplete")
-        // Apply styles after first render then unsubscribe
-        this.setFrozen()
-        this.setStyles()
-        this.renderChildren()
-        this.tabulator.modules.frozenColumns.active = true
-        this.tabulator.modules.frozenColumns.layout()
-        this.relayout()
-        this._initializing = false
-      })
-    }
-
     // Disable frozenColumns during rendering (see https://github.com/olifolkerd/tabulator/issues/3530)
     this.tabulator.on("dataLoading", () => {
       this.tabulator.modules.frozenColumns.active = false
@@ -407,54 +442,59 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     this.tabulator.on("rowSelectionChanged", (data: any, rows: any) => this.rowSelectionChanged(data, rows))
     this.tabulator.on("rowClick", (e: any, row: any) => this.rowClicked(e, row))
     this.tabulator.on("cellEdited", (cell: any) => this.cellEdited(cell))
-    this.tabulator.on("dataFiltering", () => {
-      this.model.filters = this.tabulator.getHeaderFilters()
+    this.tabulator.on("dataFiltering", (filters: any) => {
+      this.model.filters = filters
+    })
+    this.tabulator.on("dataFiltered", (_: any, rows: any[]) => {
+      if (this._initializing)
+	return
+      // Ensure that after filtering empty scroll renders
+      if (rows.length === 0)
+	this.tabulator.rowManager.renderEmptyScroll()
+      // Ensure that after filtering the page is updated
+      this.updatePage(this.tabulator.getPage())
+    })
+    this.tabulator.on("pageLoaded", (pageno: number) => {
+      this.updatePage(pageno)
+    })
+    this.tabulator.on("dataSorting", (sorters: any[]) => {
+      const sorts = []
+      for (const s of sorters) {
+        if (s.field !== '_index')
+          sorts.push({field: s.field, dir: s.dir})
+      }
+      if (this.model.pagination !== 'remote') {
+        this._updating_sort = true
+        this.model.sorters = sorts
+        this._updating_sort = false
+      }
     })
   }
 
-  tableInit(): void {
-    // Patch the ajax request and page data parsing methods
-    const ajax = this.tabulator.modules.ajax
-    ajax.sendRequest = () => {
-      return this.requestPage(ajax.params.page, ajax.params.sort)
-    }
-    this.tabulator.modules.page._parseRemoteData = (): boolean => {
-      return false
-    }
-  }
-
-
   tableBuilt(): void {
+    this._building = false
     this.setHidden()
     this.setSelection()
 
     // For remote pagination initialize on tableBuilt
-    if (this.model.pagination) {
-      this.setMaxPage()
-      this.tabulator.setPage(this.model.page)
-      this.setStyles()
-      this.renderChildren()
-      this.tabulator.modules.frozenColumns.active = true
-      this.tabulator.modules.frozenColumns.layout()
-      setTimeout(() => this.relayout(), 10)
-      this._initializing = false
-    }
-  }*/
-
-  tableInit(view: DataTabulatorView, tabulator: any): void {
-    // Patch the ajax request and page data parsing methods
-    const ajax = tabulator.modules.ajax
-    ajax.sendRequest = () => {
-      return view.requestPage(ajax.params.page, ajax.params.sorters)
-    }
-    tabulator.modules.page._parseRemoteData = () => {}
+    this.setMaxPage()
+    this.tabulator.setPage(this.model.page)
+    this.setStyles()
+    this.renderChildren()
+    this.setGroupBy()
+    this.setFrozen()
+    this.tabulator.modules.frozenColumns.active = true
+    this.tabulator.modules.frozenColumns.layout()
+    this._initializing = false
   }
 
   relayout(): void {
     if (this._relayouting)
       return
     this._relayouting = true
-    this.tabulator.rowManager.adjustTableSize()
+    if (this.tabulator.rowManager.renderer) {
+      this.tabulator.rowManager.adjustTableSize()
+    }
     this.update_layout()
     this.compute_layout()
     if (this.root !== this) {
@@ -475,14 +515,14 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     return new Promise((resolve: any, reject: any) => {
       try {
         if (page != null && sorters != null) {
-	  this._updating_sort = true
-	  const sorts = []
-	  for (const s of sorters) {
-	    if (s.field !== '_index')
-	      sorts.push({field: s.field, dir: s.dir})
-	  }
+          this._updating_sort = true
+          const sorts = []
+          for (const s of sorters) {
+            if (s.field !== '_index')
+              sorts.push({field: s.field, dir: s.dir})
+          }
           this.model.sorters = sorts
-	  this._updating_sort = false
+          this._updating_sort = false
           this._updating_page = true
           try {
             this.model.page = page || 1
@@ -516,7 +556,6 @@ export class DataTabulatorView extends PanelHTMLBoxView {
   getConfiguration(): any {
     // Only use selectable mode if explicitly requested otherwise manually handle selections
     let selectable = this.model.select_mode === 'toggle' ? true : NaN
-    const that = this
     let configuration = {
       ...this.model.configuration,
       index: "_index",
@@ -524,48 +563,16 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       movableColumns: false,
       selectable: selectable,
       columns: this.getColumns(),
-      dataSorting: (sorters: any[]) => {
-	const sorts = []
-	for (const s of sorters) {
-	  if (s.field !== '_index')
-	    sorts.push({field: s.field, dir: s.dir})
-	}
-	if (this.model.pagination !== 'remote') {
-	  this._updating_sort = true
-	  this.model.sorters = sorts
-	  this._updating_sort = false
-	}
-      },
       initialSort: this.sorters,
       layout: this.getLayout(),
-      pagination: this.model.pagination,
+      pagination: this.model.pagination != null,
+      paginationMode: this.model.pagination,
       paginationSize: this.model.page_size,
-      paginationInitialPage: 1,
-      tableBuilding: function() { that.tableInit(that, this) },
-      renderComplete: () => this.renderComplete(),
-      rowSelectionChanged: (data: any, rows: any) => this.rowSelectionChanged(data, rows),
-      rowClick: (e: any, row: any) => this.rowClicked(e, row),
-      cellEdited: (cell: any) => this.cellEdited(cell),
-      selectableCheck: (row: any) => {
-        const selectable = this.model.selectable_rows
-        return (selectable == null) || (selectable.indexOf(row._row.data._index) >= 0)
-      },
-      tooltips: (cell: any) => {
-        return  cell.getColumn().getField() + ": " + cell.getValue();
-      },
-      scrollVertical: debounce(() => {
-        this._lastVerticalScrollbarTopPosition = this.tabulator.rowManager.element.scrollTop;
-        this.setStyles()
-      }, 50, false),
-      rowFormatter: (row: any) => this._render_row(row),
-      dataFiltering: () => {
-        if (this.tabulator != null)
-          this.model.filters = this.tabulator.getHeaderFilters()
-      }
+      paginationInitialPage: 1
     }
     if (this.model.pagination === "remote") {
       configuration['ajaxURL'] = "http://panel.pyviz.org"
-      configuration['ajaxSorting'] = true
+      configuration['sortMode'] = "remote"
     }
     const cds: any = this.model.source;
     let data: any[]
@@ -613,7 +620,7 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     let viewEl = rowEl.children[rowEl.children.length-1]
     if (viewEl.className === 'bk') {
       if (viewEl.children.length)
-	return
+        return
     } else {
       viewEl = null
     }
@@ -629,7 +636,7 @@ export class DataTabulatorView extends PanelHTMLBoxView {
 
   _expand_render(cell: any): string {
     const index = cell._cell.row.data._index
-    const icon = this.model.expanded.indexOf(index) < 0 ? "\u25ba" : "\u25bc"
+    const icon = this.model.expanded.indexOf(index) < 0 ? "►" : "▼"
     return "<i>" + icon + "</i>"
   }
 
@@ -642,15 +649,13 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     else {
       const removed = expanded.splice(exp_index, 1)[0]
       if (removed in this.model.children) {
-	const model = this.model.children[removed]
-	const view = this._child_views.get(model)
-	if (view !== undefined && view.el != null)
-	  undisplay(view.el)
+        const model = this.model.children[removed]
+        const view = this._child_views.get(model)
+        if (view !== undefined && view.el != null)
+          undisplay(view.el)
       }
     }
     this.model.expanded = expanded
-    const icon = expanded.indexOf(index) < 0 ? "\u25ba" : "\u25bc"
-    cell._cell.element.innerText = icon
     if (expanded.indexOf(index) < 0)
       return
     let ready = true
@@ -745,6 +750,9 @@ export class DataTabulatorView extends PanelHTMLBoxView {
           }
         }
       }
+      if (tab_column.sorter == 'timestamp') {
+        tab_column.sorter = timestampSorter
+      }
 
       const editor: any = column.editor
       const ctype = editor.type
@@ -756,8 +764,8 @@ export class DataTabulatorView extends PanelHTMLBoxView {
         }
       } else if (ctype === "StringEditor") {
         if (editor.completions.length > 0) {
-          tab_column.editor = "autocomplete"
-          tab_column.editorParams = {values: editor.completions}
+          tab_column.editor = "list"
+          tab_column.editorParams = {values: editor.completions, autocomplete:true, listOnEmpty:true}
       } else
           tab_column.editor = "input"
       } else if (ctype === "TextEditor")
@@ -770,7 +778,7 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       } else if (ctype === "DateEditor") {
         tab_column.editor = dateEditor
       } else if (ctype === "SelectEditor") {
-        tab_column.editor = "select"
+        tab_column.editor = "list"
         tab_column.editorParams = {values: editor.options}
       } else if (editor != null && editor.default_view != null) {
         tab_column.editor = (cell: any, onRendered: any, success: any, cancel: any) => {
@@ -786,27 +794,27 @@ export class DataTabulatorView extends PanelHTMLBoxView {
         }
       }
       for (const sort of this.model.sorters) {
-	if (tab_column.field === sort.field)
-	  tab_column.headerSortStartingDir = sort.dir
+        if (tab_column.field === sort.field)
+          tab_column.headerSortStartingDir = sort.dir
       }
       tab_column.cellClick = (_: any, cell: any) => {
-	const index = cell._cell.row.data._index;
-	this.model.trigger_event(new CellClickEvent(column.field, index))
+        const index = cell.getData()._index
+        this.model.trigger_event(new CellClickEvent(column.field, index))
       }
       if (config_columns == null)
         columns.push(tab_column)
     }
     for (const col in this.model.buttons) {
       const button_formatter = () => {
-	return this.model.buttons[col];
+        return this.model.buttons[col];
       };
       const button_column = {
-	formatter: button_formatter,
-	hozAlign: "center",
-	cellClick: (_: any, cell: any) => {
-	  const index = cell._cell.row.data._index;
-	  this.model.trigger_event(new CellClickEvent(col, index))
-	}
+        formatter: button_formatter,
+        hozAlign: "center",
+        cellClick: (_: any, cell: any) => {
+          const index = cell.getData()._index
+          this.model.trigger_event(new CellClickEvent(col, index))
+        }
       }
       columns.push(button_column)
     }
@@ -882,6 +890,14 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       this.tabulator.getRow(row).freeze()
   }
 
+  updatePage(pageno: number): void {
+    if (this.model.pagination === 'local' && this.model.page !== pageno) {
+      this._updating_page = true
+      this.model.page = pageno
+      this._updating_page = false
+    }
+  }
+
   setGroupBy(): void {
     if (this.model.groupby.length == 0) {
       this.tabulator.setGroupBy(false)
@@ -895,6 +911,8 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       }
       return groups.join(', ')
     }
+    // Need to call it twice, see https://github.com/olifolkerd/tabulator/issues/3666
+    this.tabulator.setGroupBy(groupby)
     this.tabulator.setGroupBy(groupby)
   }
 
@@ -906,10 +924,22 @@ export class DataTabulatorView extends PanelHTMLBoxView {
 
   setCSS(): boolean {
     let theme: string
-    if (this.model.theme == "default")
+    let theme_: string
+    if (this.model.theme == "default") {
       theme = "tabulator"
-    else
-      theme = "tabulator_" + this.model.theme
+    }
+    else {
+      if (this.model.theme == "bootstrap") {
+        theme_ = "bootstrap3"
+      }
+      else if (this.model.theme == "semantic-ui") {
+        theme_ = "semanticui"
+      }
+      else {
+        theme_ = this.model.theme
+      }
+      theme = "tabulator_" + theme_
+    }
     const css = this.model.theme_url + theme + ".min.css"
 
     let old_node: any = null
@@ -941,8 +971,10 @@ export class DataTabulatorView extends PanelHTMLBoxView {
     css_node.href = css
 
     css_node.onload = () => {
-      this.render()
-      this.relayout()
+      if (!this._building) {
+	this.render()
+	this.relayout()
+      }
     }
     parent_node.appendChild(css_node)
     return true
@@ -973,7 +1005,7 @@ export class DataTabulatorView extends PanelHTMLBoxView {
           else
             [prop, value] = s.split(':')
           element.style.setProperty(prop, value.trimLeft())
-	  this._applied_styles = true
+          this._applied_styles = true
         }
       }
     }
@@ -1026,11 +1058,11 @@ export class DataTabulatorView extends PanelHTMLBoxView {
   rowClicked(e: any, row: any) {
     if (
         this._selection_updating ||
-	this._initializing ||
-	(typeof this.model.select_mode) === 'string' ||
-	this.model.select_mode === false ||  // selection disabled
-	this.model.configuration.dataTree || // dataTree does not support selection
-	e.srcElement?.innerText === "►"      // expand button
+        this._initializing ||
+        (typeof this.model.select_mode) === 'string' ||
+        this.model.select_mode === false ||  // selection disabled
+        this.model.configuration.dataTree || // dataTree does not support selection
+        e.srcElement?.innerText === "►"      // expand button
     )
       return
     let indices: number[] = []
@@ -1079,10 +1111,10 @@ export class DataTabulatorView extends PanelHTMLBoxView {
   rowSelectionChanged(data: any, _: any): void {
     if (
         this._selection_updating ||
-	this._initializing ||
-	(typeof this.model.select_mode) === 'boolean' ||
-	(typeof this.model.select_mode) === 'number' ||
-	this.model.configuration.dataTree
+        this._initializing ||
+        (typeof this.model.select_mode) === 'boolean' ||
+        (typeof this.model.select_mode) === 'number' ||
+        this.model.configuration.dataTree
     )
       return
     const indices: number[] = data.map((row: any) => row._index)
@@ -1094,8 +1126,8 @@ export class DataTabulatorView extends PanelHTMLBoxView {
 
   cellEdited(cell: any): void {
     const field = cell._cell.column.field;
-    const index = cell._cell.row.data._index;
-    const value = cell._cell.value;
+    const index = cell.getData()._index
+    const value = cell._cell.value
     this._tabulator_cell_updating = true
     comm_settings.debounce = false
     try {
@@ -1105,7 +1137,6 @@ export class DataTabulatorView extends PanelHTMLBoxView {
       this._tabulator_cell_updating = false
     }
     this.model.trigger_event(new TableEditEvent(field, index))
-    this.setSorters()
     this.tabulator.scrollToRow(index, "top", false)
   }
 }
@@ -1187,7 +1218,7 @@ export class DataTabulator extends HTMLBox {
       sorters:        [ Array(Any),              [] ],
       styles:         [ Any,                     {} ],
       theme:          [ String,            "simple" ],
-      theme_url:      [ String, "https://unpkg.com/tabulator-tables@4.9.3/dist/css/"]
+      theme_url:      [ String, "https://unpkg.com/tabulator-tables@5.3.2/dist/css/"]
     }))
   }
 }
