@@ -2,10 +2,9 @@
 Custom bokeh Widget models.
 """
 from bokeh.core.enums import ButtonType
-
 from bokeh.core.properties import (
-    Any, Bool, Dict, Either, Enum, Float, Int, List, Nullable,
-    Override, String, Tuple
+    Any, Bool, Dict, Either, Enum, Float, Int, List, Nullable, Override,
+    String, Tuple,
 )
 from bokeh.models.layouts import HTMLBox
 from bokeh.models.widgets import InputWidget, Select, Widget
@@ -22,6 +21,8 @@ class Player(Widget):
 
     value = Int(0, help="Current value of the player app")
 
+    value_throttled = Int(0, help="Current throttled value of the player app")
+
     step = Int(1, help="Number of steps to advance the player by.")
 
     interval = Int(500, help="Interval between updates")
