@@ -8,16 +8,18 @@ import warnings
 
 from functools import partial
 from typing import (
-    TYPE_CHECKING, Any, Callable, List, Optional, TypeVar, Type
+    TYPE_CHECKING, Any, Callable, ClassVar, List, Optional, Type, TypeVar,
 )
 
 import param
 
 from bokeh.models.layouts import (
-    GridBox as _BkGridBox, Tabs as _BkTabs, Panel as _BkPanel
+    GridBox as _BkGridBox, Panel as _BkPanel, Tabs as _BkTabs,
 )
 
-from ..io import init_doc, push, state, unlocked
+from ..io import (
+    init_doc, push, state, unlocked,
+)
 from ..layout import Panel, Row
 from ..links import Link
 from ..models import ReactiveHTML as _BkReactiveHTML
@@ -119,19 +121,19 @@ class PaneBase(Reactive):
     # numerical priority is selected. The default is an intermediate value.
     # If set to None, applies method will be called to get a priority
     # value for a specific object type.
-    priority: float | bool | None = 0.5
+    priority: ClassVar[float | bool | None] = 0.5
 
     # Whether applies requires full set of keywords
-    _applies_kw = False
+    _applies_kw: ClassVar[bool] = False
 
     # Whether the Pane layout can be safely unpacked
-    _unpack: bool = True
+    _unpack: ClassVar[bool] = True
 
     # Declares whether Pane supports updates to the Bokeh model
-    _updates: bool = False
+    _updates: ClassVar[bool] = False
 
     # List of parameters that trigger a rerender of the Bokeh model
-    _rerender_params: List[str] = ['object']
+    _rerender_params: ClassVar[List[str]] = ['object']
 
     __abstract = True
 
@@ -177,7 +179,7 @@ class PaneBase(Reactive):
         return [p for p in self.param if p not in ignored_params]
 
     def _update_object(
-        self, ref: str, doc: 'Document', root: 'Model', parent: 'Model', comm: Optional['Comm']
+        self, ref: str, doc: 'Document', root: Model, parent: Model, comm: Optional[Comm]
     ) -> None:
         old_model = self._models[ref][0]
         if self._updates:
@@ -207,11 +209,11 @@ class PaneBase(Reactive):
                 else:
                     index = parent.children.index(old_model)
             except ValueError:
-                self.warning('%s pane model %s could not be replaced '
-                             'with new model %s, ensure that the '
-                             'parent is not modified at the same '
-                             'time the panel is being updated.' %
-                             (type(self).__name__, old_model, new_model))
+                self.param.warning(
+                    f'{type(self).__name__} pane model {old_model!r} could not be '
+                    f'replaced with new model {new_model!r}, ensure that the parent '
+                    'is not modified at the same time the panel is being updated.'
+                )
             else:
                 if isinstance(parent, _BkReactiveHTML):
                     parent.children[node] = new_models
@@ -244,7 +246,7 @@ class PaneBase(Reactive):
                 else:
                     cb()
 
-    def _update(self, ref: Optional[str] = None, model: Optional['Model'] = None) -> None:
+    def _update(self, ref: str, model: Model) -> None:
         """
         If _updates=True this method is used to update an existing
         Bokeh model instead of replacing the model entirely. The
@@ -288,9 +290,9 @@ class PaneBase(Reactive):
         return type(self)(object, **params)
 
     def get_root(
-        self, doc: Optional['Document'] = None, comm: Optional['Comm'] = None,
+        self, doc: Optional[Document] = None, comm: Optional[Comm] = None,
         preprocess: bool = True
-    ) -> 'Model':
+    ) -> Model:
         """
         Returns the root model and applies pre-processing hooks
 
@@ -384,7 +386,7 @@ class ReplacementPane(PaneBase):
         self._kwargs =  {p: params.pop(p) for p in list(params)
                          if p not in self.param}
         super().__init__(object, **params)
-        self._pane = Pane(None)
+        self._pane = panel(None)
         self._internal = True
         self._inner_layout = Row(self._pane, **{k: v for k, v in params.items() if k in Row.param})
         self.param.watch(self._update_inner_layout, list(Layoutable.param))
@@ -458,9 +460,9 @@ class ReplacementPane(PaneBase):
         self._internal = internal
 
     def _get_model(
-        self, doc: 'Document', root: Optional['Model'] = None,
-        parent: Optional['Model'] = None, comm: Optional['Comm'] = None
-    ) -> 'Model':
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         if root:
             ref = root.ref['id']
             if ref in self._models:
@@ -471,7 +473,7 @@ class ReplacementPane(PaneBase):
         self._models[ref] = (model, parent)
         return model
 
-    def _cleanup(self, root: 'Model' | None = None) -> None:
+    def _cleanup(self, root: Model | None = None) -> None:
         self._inner_layout._cleanup(root)
         super()._cleanup(root)
 
