@@ -1,5 +1,11 @@
-import sys
+from __future__ import annotations
+
 import json
+import sys
+
+from typing import (
+    TYPE_CHECKING, Any, ClassVar, List, Mapping, Optional,
+)
 
 import param
 
@@ -8,6 +14,11 @@ from pyviz_comms import JupyterComm
 from ..util import lazy_load
 from .base import PaneBase
 
+if TYPE_CHECKING:
+    from bokeh.document import Document
+    from bokeh.model import Model
+    from pyviz_comms import Comm
+
 
 class ECharts(PaneBase):
     """
@@ -30,16 +41,16 @@ class ECharts(PaneBase):
     theme = param.ObjectSelector(default="default", objects=["default", "light", "dark"], doc="""
        Theme to apply to plots.""")
 
-    priority = None
+    priority: ClassVar[float | bool | None] = None
 
-    _rename = {"object": "data"}
+    _rename: ClassVar[Mapping[str, str | None]] = {"object": "data"}
 
-    _rerender_params = []
+    _rerender_params: ClassVar[List[str]] = []
 
-    _updates = True
+    _updates: ClassVar[bool] = True
 
     @classmethod
-    def applies(cls, obj, **params):
+    def applies(cls, obj: Any, **params) -> float | bool | None:
         if isinstance(obj, dict):
             return 0
         elif "pyecharts." in repr(obj.__class__):
@@ -56,7 +67,10 @@ class ECharts(PaneBase):
         else:
             props['sizing_mode'] = 'fixed'
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         ECharts = lazy_load('panel.models.echarts', 'ECharts', isinstance(comm, JupyterComm), root)
         props = self._get_echart_dict(self.object)
         props.update(self._process_param_change(self._init_params()))
