@@ -1,5 +1,11 @@
+from __future__ import annotations
+
 import os
 
+from typing import (
+    TYPE_CHECKING, Any, ClassVar, Optional,
+)
+
 import param
 
 from pyviz_comms import JupyterComm
@@ -8,6 +14,11 @@ from ..config import config
 from ..models import IPyWidget as _BkIPyWidget
 from .base import PaneBase
 
+if TYPE_CHECKING:
+    from bokeh.document import Document
+    from bokeh.model import Model
+    from pyviz_comms import Comm
+
 
 class IPyWidget(PaneBase):
     """
@@ -25,10 +36,10 @@ class IPyWidget(PaneBase):
     >>> IPyWidget(some_ipywidget)
     """
 
-    priority = 0.6
+    priority: ClassVar[float | bool | None] = 0.6
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return (hasattr(obj, 'traits') and hasattr(obj, 'get_manager_state') and hasattr(obj, 'comm'))
 
     def _get_ipywidget(self, obj, doc, root, comm, **kwargs):
@@ -36,7 +47,9 @@ class IPyWidget(PaneBase):
             IPyWidget = _BkIPyWidget
         else:
             import ipykernel
+
             from ipywidgets_bokeh.widget import IPyWidget
+
             from ..io.ipywidget import PanelKernel
 
             # Patch font-awesome CSS
@@ -46,14 +59,19 @@ class IPyWidget(PaneBase):
 
             if not isinstance(ipykernel.kernelbase.Kernel._instance, PanelKernel):
                 kernel = PanelKernel(document=doc, key=str(id(doc)).encode('utf-8'))
-                for w in obj.widgets.values():
+                # Support ipywidgets >=8.0 and <8.0
+                widgets = (obj._active_widgets if hasattr(obj, '_active_widgets') else obj.widgets).values()
+                for w in widgets:
                     w.comm.kernel = kernel
                     w.comm.open()
 
         model = IPyWidget(widget=obj, **kwargs)
         return model
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         if root is None:
             return self.get_root(doc, comm)
         kwargs = self._process_param_change(self._init_params())
@@ -68,8 +86,8 @@ class IPyLeaflet(IPyWidget):
         'fixed', 'stretch_width', 'stretch_height', 'stretch_both',
         'scale_width', 'scale_height', 'scale_both', None])
 
-    priority = 0.7
+    priority: float | bool | None = 0.7
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return IPyWidget.applies(obj) and obj._view_module == 'jupyter-leaflet'
