@@ -1,15 +1,21 @@
 """
 Pane class which render plots from different libraries
 """
+from __future__ import annotations
+
 import sys
 
-from io import BytesIO
-
 from contextlib import contextmanager
+from io import BytesIO
+from typing import (
+    TYPE_CHECKING, Any, ClassVar, Mapping, Optional,
+)
 
 import param
 
-from bokeh.models import CustomJS, LayoutDOM, Model, Spacer as BkSpacer
+from bokeh.models import (
+    CustomJS, LayoutDOM, Model, Spacer as BkSpacer,
+)
 from bokeh.themes import Theme
 
 from ..io import remove_root
@@ -17,9 +23,13 @@ from ..io.notebook import push
 from ..util import escape
 from ..viewable import Layoutable
 from .base import PaneBase
+from .image import PNG
 from .ipywidget import IPyWidget
 from .markup import HTML
-from .image import PNG
+
+if TYPE_CHECKING:
+    from bokeh.document import Document
+    from pyviz_comms import Comm
 
 FOLIUM_BEFORE = '<div style="width:100%;"><div style="position:relative;width:100%;height:0;padding-bottom:60%;">'
 FOLIUM_AFTER = '<div style="width:100%;height:100%"><div style="position:relative;width:100%;height:100%;padding-bottom:0%;">'
@@ -64,12 +74,12 @@ class Bokeh(PaneBase):
     theme = param.ClassSelector(default=None, class_=(Theme, str), doc="""
         Bokeh theme to apply to the plot.""")
 
-    priority = 0.8
+    priority: ClassVar[float | bool | None] = 0.8
 
-    _rename = {'autodispatch': None, 'theme': None}
+    _rename: ClassVar[Mapping[str, str | None]] = {'autodispatch': None, 'theme': None}
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return isinstance(obj, LayoutDOM)
 
     @classmethod
@@ -102,7 +112,10 @@ class Bokeh(PaneBase):
                     for cb in cbs
                 ]
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         if root is None:
             return self.get_root(doc, comm)
 
@@ -157,19 +170,24 @@ class Matplotlib(PNG, IPyWidget):
     dpi = param.Integer(default=144, bounds=(1, None), doc="""
         Scales the dpi of the matplotlib figure.""")
 
+    high_dpi = param.Boolean(default=True, doc="""
+        Whether to optimize output for high-dpi displays.""")
+
     interactive = param.Boolean(default=False, constant=True, doc="""
-    """)
+        Whether to render interactive matplotlib plot with ipympl.""")
 
     tight = param.Boolean(default=False, doc="""
         Automatically adjust the figure size to fit the
         subplots and other artist elements.""")
 
-    _rename = {'object': 'text', 'interactive': None, 'dpi': None,  'tight': None}
+    _rename: ClassVar[Mapping[str, str | None]] = {
+        'object': 'text', 'interactive': None, 'dpi': None,  'tight': None, 'high_dpi': None
+    }
 
     _rerender_params = PNG._rerender_params + ['object', 'dpi', 'tight']
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         if 'matplotlib' not in sys.modules:
             return False
         from matplotlib.figure import Figure
@@ -187,7 +205,7 @@ class Matplotlib(PNG, IPyWidget):
         import matplotlib.backends
         old_backend = getattr(matplotlib.backends, 'backend', 'agg')
 
-        from ipympl.backend_nbagg import FigureManager, Canvas, is_interactive
+        from ipympl.backend_nbagg import Canvas, FigureManager, is_interactive
         from matplotlib._pylab_helpers import Gcf
 
         matplotlib.use(old_backend)
@@ -205,28 +223,41 @@ class Matplotlib(PNG, IPyWidget):
         canvas.mpl_connect('close_event', closer)
         return manager
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _update_dimensions(self):
+        w, h = self.object.get_size_inches()
+        dpi = self.dpi / 2. if self.high_dpi else self.dpi
+        with param.discard_events(self):
+            self.width = self.width or int(dpi * w)
+            self.height = self.height or int(dpi * h)
+
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
+        self._update_dimensions()
         if not self.interactive:
-            return PNG._get_model(self, doc, root, parent, comm)
+            model = PNG._get_model(self, doc, root, parent, comm)
+            return model
         self.object.set_dpi(self.dpi)
         manager = self._get_widget(self.object)
         props = self._process_param_change(self._init_params())
         kwargs = {k: v for k, v in props.items()
                   if k not in self._rerender_params+['interactive']}
-        w, h = self.object.get_size_inches()
-        kwargs['width'] = self.width or int(self.dpi * w)
-        kwargs['height'] = self.height or int(self.dpi * h)
+        kwargs['width'] = self.width
+        kwargs['height'] = self.height
         kwargs['sizing_mode'] = self.sizing_mode
-        model = self._get_ipywidget(manager.canvas, doc, root, comm,
-                                    **kwargs)
+        model = self._get_ipywidget(
+            manager.canvas, doc, root, comm, **kwargs
+        )
         if root is None:
             root = model
         self._models[root.ref['id']] = (model, parent)
         self._managers[root.ref['id']] = manager
         return model
 
-    def _update(self, ref=None, model=None):
+    def _update(self, ref: str, model: Model) -> None:
         if not self.interactive:
+            self._update_dimensions()
             model.update(**self._get_properties())
             return
         manager = self._managers[ref]
@@ -268,12 +299,12 @@ class RGGPlot(PNG):
     _rerender_params = PNG._rerender_params + ['object', 'dpi', 'width', 'height']
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return type(obj).__name__ == 'GGPlot' and hasattr(obj, 'r_repr')
 
     def _img(self):
-        from rpy2.robjects.lib import grdevices
         from rpy2 import robjects
+        from rpy2.robjects.lib import grdevices
         with grdevices.render_to_bytesio(grdevices.png,
                  type="cairo-png", width=self.width, height=self.height,
                  res=self.dpi, antialias="subpixel") as b:
@@ -289,10 +320,10 @@ class YT(HTML):
     provide additional space.
     """
 
-    priority = 0.5
+    priority: ClassVar[float | bool | None] = 0.5
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: bool) -> float | bool | None:
         return (getattr(obj, '__module__', '').startswith('yt.') and
                 hasattr(obj, "plots") and
                 hasattr(obj, "_repr_html_"))
@@ -326,10 +357,10 @@ class Folium(HTML):
         'fixed', 'stretch_width', 'stretch_height', 'stretch_both',
         'scale_width', 'scale_height', 'scale_both', None])
 
-    priority = 0.6
+    priority: ClassVar[float | bool | None] = 0.6
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return (getattr(obj, '__module__', '').startswith('folium.') and
                 hasattr(obj, "_repr_html_"))
 
