@@ -2,43 +2,56 @@
 Defines the Param pane which converts Parameterized classes into a
 set of widgets.
 """
+from __future__ import annotations
+
 import inspect
 import itertools
 import json
 import os
+import sys
 import types
 
-from collections.abc import Callable
 from collections import OrderedDict, defaultdict, namedtuple
+from collections.abc import Callable
 from contextlib import contextmanager
 from functools import partial
-from packaging.version import Version
+from typing import (
+    TYPE_CHECKING, Any, ClassVar, List, Mapping, Optional, Type,
+)
 
 import param
 
+from packaging.version import Version
 from param.parameterized import classlist, discard_events
 
+from .config import config
 from .io import init_doc, state
-from .layout import Column, Panel, Row, Spacer, Tabs
-from .pane.base import PaneBase, ReplacementPane
-from .util import (
-    abbreviated_repr, bokeh_version, classproperty, full_groupby,
-    fullpath, get_method_owner, is_parameterized, param_name,
-    recursive_parameterized
+from .layout import (
+    Column, Panel, Row, Spacer, Tabs,
 )
+from .pane.base import PaneBase, ReplacementPane
 from .reactive import Reactive
+from .util import (
+    abbreviated_repr, bokeh_version, classproperty, full_groupby, fullpath,
+    get_method_owner, is_parameterized, param_name, recursive_parameterized,
+)
 from .viewable import Layoutable, Viewable
 from .widgets import (
     ArrayInput, Button, Checkbox, ColorPicker, DataFrame, DatePicker,
-    DatetimeInput, DatetimeRangeSlider, DateRangeSlider,
-    DiscreteSlider, FileSelector, FloatSlider, IntInput, IntSlider,
-    LiteralInput, MultiSelect, RangeSlider, Select, FloatInput,
-    StaticText, TextInput, Toggle, Widget
+    DateRangeSlider, DatetimeInput, DatetimeRangeSlider, DiscreteSlider,
+    FileSelector, FloatInput, FloatSlider, IntInput, IntSlider, LiteralInput,
+    MultiSelect, RangeSlider, Select, StaticText, Tabulator, TextInput, Toggle,
+    Widget,
 )
 from .widgets.button import _ButtonBase
 
+if TYPE_CHECKING:
+    from bokeh.document import Document
+    from bokeh.model import Model
+    from pyviz_comms import Comm
 
-def SingleFileSelector(pobj):
+
+def SingleFileSelector(pobj: param.Parameter) -> Type[Widget]:
     """
     Determines whether to use a TextInput or Select widget for FileSelector
     """
@@ -48,7 +61,7 @@ def SingleFileSelector(pobj):
         return TextInput
 
 
-def LiteralInputTyped(pobj):
+def LiteralInputTyped(pobj: param.Parameter) -> Type[Widget]:
     if isinstance(pobj, param.Tuple):
         return type(str('TupleInput'), (LiteralInput,), {'type': tuple})
     elif isinstance(pobj, param.Number):
@@ -60,6 +73,13 @@ def LiteralInputTyped(pobj):
     return LiteralInput
 
 
+def DataFrameWidget(pobj: param.DataFrame) -> Type[Widget]:
+    if 'panel.models.tabulator' in sys.modules:
+        return Tabulator
+    else:
+        return DataFrame
+
+
 @contextmanager
 def set_values(*parameterizeds, **param_values):
     """
@@ -151,7 +171,7 @@ class Param(PaneBase):
         Dictionary of widget overrides, mapping from parameter name
         to widget class.""")
 
-    mapping = {
+    mapping: ClassVar[Mapping[param.Parameter, Widget | Callable[[param.Parameter], Widget]]] = {
         param.Action:            Button,
         param.Array:             ArrayInput,
         param.Boolean:           Checkbox,
@@ -160,7 +180,7 @@ class Param(PaneBase):
         param.Date:              DatetimeInput,
         param.DateRange:         DateRangeSlider,
         param.CalendarDateRange: DateRangeSlider,
-        param.DataFrame:         DataFrame,
+        param.DataFrame:         DataFrameWidget,
         param.Dict:              LiteralInputTyped,
         param.FileSelector:      SingleFileSelector,
         param.Filename:          TextInput,
@@ -183,11 +203,11 @@ class Param(PaneBase):
     if bokeh_version >= Version('2.4.3'):
         mapping[param.DateRange] = DatetimeRangeSlider
 
-    priority = 0.1
+    priority: ClassVar[float | bool | None] = 0.1
 
-    _unpack = True
+    _unpack: ClassVar[bool] = True
 
-    _rerender_params = []
+    _rerender_params: ClassVar[List[str]] = []
 
     def __init__(self, object=None, **params):
         if isinstance(object, param.Parameter):
@@ -412,12 +432,12 @@ class Param(PaneBase):
             kw['visible'] = not p_obj.constant
 
         value = getattr(self.object, p_name)
-        if value is not None:
+        allow_None = p_obj.allow_None or False
+        if isinstance(widget_class, type) and issubclass(widget_class, Widget):
+            allow_None &= widget_class.param.value.allow_None
+        if value is not None or allow_None:
             kw['value'] = value
 
-        # Update kwargs
-        kw.update(kw_widget)
-
         if hasattr(p_obj, 'get_range'):
             options = p_obj.get_range()
             # This applies to widgets whose `options` Parameter is a List type,
@@ -446,6 +466,13 @@ class Param(PaneBase):
                         widget_class = LiteralInput
             if hasattr(widget_class, 'step') and getattr(p_obj, 'step', None):
                 kw['step'] = p_obj.step
+            if hasattr(widget_class, 'fixed_start') and getattr(p_obj, 'bounds', None):
+                kw['fixed_start'] = p_obj.bounds[0]
+            if hasattr(widget_class, 'fixed_end') and getattr(p_obj, 'bounds', None):
+                kw['fixed_end'] = p_obj.bounds[1]
+
+        # Update kwargs
+        kw.update(kw_widget)
 
         kwargs = {k: v for k, v in kw.items() if k in widget_class.param}
 
@@ -478,6 +505,8 @@ class Param(PaneBase):
             def action(change):
                 value(self.object)
             watcher = widget.param.watch(action, 'clicks')
+        elif kw_widget.get('onkeyup', False) and hasattr(widget, 'value_input'):
+            watcher = widget.param.watch(link_widget, 'value_input')
         elif kw_widget.get('throttled', False) and hasattr(widget, 'value_throttled'):
             watcher = widget.param.watch(link_widget, 'value_throttled')
         else:
@@ -647,12 +676,15 @@ class Param(PaneBase):
         widgets += [(pname, self.widget(pname)) for pname in self._ordered_params]
         return OrderedDict(widgets)
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         model = self.layout._get_model(doc, root, parent, comm)
         self._models[root.ref['id']] = (model, parent)
         return model
 
-    def _cleanup(self, root):
+    def _cleanup(self, root: Model | None = None) -> None:
         self.layout._cleanup(root)
         super()._cleanup(root)
 
@@ -661,7 +693,7 @@ class Param(PaneBase):
     #----------------------------------------------------------------
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return (is_parameterized(obj) or
                 isinstance(obj, param.parameterized.Parameters) or
                 (isinstance(obj, param.Parameter) and obj.owner is not None))
@@ -694,7 +726,10 @@ class Param(PaneBase):
         """
         return super().select(selector) + self.layout.select(selector)
 
-    def get_root(self, doc=None, comm=None, preprocess=True):
+    def get_root(
+        self, doc: Optional[Document] = None, comm: Optional[Comm] = None,
+        preprocess: bool = True
+    ) -> Model:
         """
         Returns the root model and applies pre-processing hooks
 
@@ -729,6 +764,10 @@ class ParamMethod(ReplacementPane):
     return any object which itself can be rendered as a Pane.
     """
 
+    defer_load = param.Boolean(default=None, doc="""
+        Whether to defer load until after the page is rendered.
+        Can be set as parameter or by setting panel.config.defer_load.""")
+
     lazy = param.Boolean(default=False, doc="""
         Whether to lazily evaluate the contents of the object
         only when it is required for rendering.""")
@@ -737,12 +776,18 @@ class ParamMethod(ReplacementPane):
         Whether to show loading indicator while pane is updating.""")
 
     def __init__(self, object=None, **params):
+        if (
+            self.param.defer_load.default is None and
+            'defer_load' not in params and config.defer_load
+        ):
+            params['defer_load'] = config.defer_load
         super().__init__(object, **params)
-        self._evaled = not self.lazy
+        self._evaled = not (self.lazy or self.defer_load)
         self._link_object_params()
         if object is not None:
             self._validate_object()
-            self._replace_pane(not self.lazy)
+            if not self.defer_load:
+                self._replace_pane()
 
     @param.depends('object', watch=True)
     def _validate_object(self):
@@ -784,20 +829,23 @@ class ParamMethod(ReplacementPane):
             self._inner_layout.loading = False
 
     def _replace_pane(self, *args, force=False):
-        self._evaled = bool(self._models) or force or not self.lazy
-        if self._evaled:
-            self._inner_layout.loading = self.loading_indicator
-            try:
-                if self.object is None:
-                    new_object = Spacer()
-                else:
-                    new_object = self.eval(self.object)
-                if inspect.isawaitable(new_object):
-                    param.parameterized.async_executor(partial(self._eval_async, new_object))
-                    return
-                self._update_inner(new_object)
-            finally:
-                self._inner_layout.loading = False
+        deferred = self.defer_load and not state.loaded
+        if not self._inner_layout.loading:
+            self._inner_layout.loading = bool(self.loading_indicator or deferred)
+        self._evaled |= force or not (self.lazy or deferred)
+        if not self._evaled:
+            return
+        try:
+            if self.object is None:
+                new_object = Spacer()
+            else:
+                new_object = self.eval(self.object)
+            if inspect.isawaitable(new_object):
+                param.parameterized.async_executor(partial(self._eval_async, new_object))
+                return
+            self._update_inner(new_object)
+        finally:
+            self._inner_layout.loading = False
 
     def _update_pane(self, *events):
         callbacks = []
@@ -853,9 +901,15 @@ class ParamMethod(ReplacementPane):
             watcher = pobj.param.watch(update_pane, ps, p.what)
             self._callbacks.append(watcher)
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         if not self._evaled:
-            self._replace_pane(force=True)
+            deferred = self.defer_load and not state.loaded
+            if deferred:
+                state.onload(partial(self._replace_pane, force=True))
+            self._replace_pane(force=not deferred)
         return super()._get_model(doc, root, parent, comm)
 
     #----------------------------------------------------------------
@@ -863,7 +917,7 @@ class ParamMethod(ReplacementPane):
     #----------------------------------------------------------------
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any) -> float | bool | None:
         return inspect.ismethod(obj) and isinstance(get_method_owner(obj), param.Parameterized)
 
 
@@ -877,12 +931,14 @@ class ParamFunction(ParamMethod):
     a widget to some other output.
     """
 
-    priority = 0.6
+    priority: ClassVar[float | bool | None] = 0.6
+
+    _applies_kw: ClassVar[bool] = True
 
     def _link_object_params(self):
         deps = getattr(self.object, '_dinfo', {})
         dep_params = list(deps.get('dependencies', [])) + list(deps.get('kw', {}).values())
-        if not dep_params:
+        if not dep_params and not self.lazy and not self.defer_load:
             fn = getattr(self.object, '__bound_function__', self.object)
             fn_name = getattr(fn, '__name__', repr(self.object))
             self.param.warning(
@@ -913,10 +969,12 @@ class ParamFunction(ParamMethod):
     #----------------------------------------------------------------
 
     @classmethod
-    def applies(cls, obj):
+    def applies(cls, obj: Any, **kwargs) -> float | bool | None:
         if isinstance(obj, types.FunctionType):
             if hasattr(obj, '_dinfo'):
                 return True
+            if kwargs.get('defer_load') or (cls.param.defer_load.default is None and config.defer_load):
+                return True
             return None
         return False
 
