@@ -16,8 +16,8 @@ from collections import Counter, defaultdict, namedtuple
 from functools import partial
 from pprint import pformat
 from typing import (
-    TYPE_CHECKING, Any, Callable, Dict, List, Mapping, Optional,
-    Set, Tuple, Type, Union
+    TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Mapping, Optional, Set,
+    Tuple, Type, Union,
 )
 
 import bleach
@@ -26,6 +26,7 @@ import param
 
 from bokeh.core.property.descriptors import UnsetValueError
 from bokeh.model import DataModel
+from packaging.version import Version
 from param.parameterized import ParameterizedMetaclass, Watcher
 
 from .io.document import unlocked
@@ -33,7 +34,7 @@ from .io.model import hold
 from .io.notebook import push
 from .io.state import set_curdoc, state
 from .models.reactive_html import (
-    DOMEvent, ReactiveHTML as _BkReactiveHTML, ReactiveHTMLParser
+    DOMEvent, ReactiveHTML as _BkReactiveHTML, ReactiveHTMLParser,
 )
 from .util import edit_readonly, escape, updating
 from .viewable import Layoutable, Renderable, Viewable
@@ -72,28 +73,28 @@ class Syncable(Renderable):
     """
 
     # Timeout if a notebook comm message is swallowed
-    _timeout: int = 20000
+    _timeout: ClassVar[int] = 20000
 
     # Timeout before the first event is processed
-    _debounce: int = 50
+    _debounce: ClassVar[int] = 50
 
     # Property changes which should not be debounced
-    _priority_changes: List[str] = []
+    _priority_changes: ClassVar[List[str]] = []
 
     # Any parameters that require manual updates handling for the models
     # e.g. parameters which affect some sub-model
-    _manual_params: List[str] = []
+    _manual_params: ClassVar[List[str]] = []
 
     # Mapping from parameter name to bokeh model property name
-    _rename: Mapping[str, str | None] = {}
+    _rename: ClassVar[Mapping[str, str | None]] = {}
 
     # Allows defining a mapping from model property name to a JS code
     # snippet that transforms the object before serialization
-    _js_transforms: Mapping[str, str] = {}
+    _js_transforms: ClassVar[Mapping[str, str]] = {}
 
     # Transforms from input value to bokeh property value
-    _source_transforms: Mapping[str, str | None] = {}
-    _target_transforms: Mapping[str, str | None] = {}
+    _source_transforms: ClassVar[Mapping[str, str | None]] = {}
+    _target_transforms: ClassVar[Mapping[str, str | None]] = {}
 
     __abstract = True
 
@@ -178,8 +179,8 @@ class Syncable(Renderable):
             self._callbacks.append(watcher)
 
     def _link_props(
-        self, model: 'Model', properties: List[str] | List[Tuple[str, str]],
-        doc: 'Document', root: 'Model', comm: Optional['Comm'] = None
+        self, model: Model, properties: List[str] | List[Tuple[str, str]],
+        doc: Document, root: Model, comm: Optional[Comm] = None
     ) -> None:
         from .config import config
         ref = root.ref['id']
@@ -202,8 +203,8 @@ class Syncable(Renderable):
                 m.on_change(p, partial(self._server_change, doc, ref, subpath))
 
     def _manual_update(
-        self, events: Tuple[param.parameterized.Event, ...], model: 'Model', doc: 'Document',
-        root: 'Model', parent: Optional['Model'], comm: Optional['Comm']
+        self, events: Tuple[param.parameterized.Event, ...], model: Model, doc: Document,
+        root: Model, parent: Optional[Model], comm: Optional[Comm]
     ) -> None:
         """
         Method for handling any manual update events, i.e. events triggered
@@ -228,8 +229,8 @@ class Syncable(Renderable):
                     cb()
 
     def _apply_update(
-        self, events: Dict[str, param.parameterized.Event], msg: Mapping[str, Any],
-        model: 'Model', ref: str
+        self, events: Dict[str, param.parameterized.Event], msg: Dict[str, Any],
+        model: Model, ref: str
     ) -> None:
         if ref not in state._views or ref in state._fake_roots:
             return
@@ -245,7 +246,7 @@ class Syncable(Renderable):
 
     def _update_model(
         self, events: Dict[str, param.parameterized.Event], msg: Dict[str, Any],
-        root: 'Model', model: 'Model', doc: 'Document', comm: Optional['Comm']
+        root: Model, model: Model, doc: Document, comm: Optional[Comm]
     ) -> None:
         ref = root.ref['id']
         self._changing[ref] = attrs = []
@@ -270,12 +271,15 @@ class Syncable(Renderable):
             elif ref in self._changing:
                 del self._changing[ref]
 
-    def _cleanup(self, root: 'Model' | None) -> None:
+    def _cleanup(self, root: Model | None) -> None:
         super()._cleanup(root)
         if root is None:
             return
         ref = root.ref['id']
-        self._models.pop(ref, None)
+        if ref in self._models:
+            model, _ = self._models.pop(ref, None)
+            model._callbacks = {}
+            model._event_callbacks = {}
         comm, client_comm = self._comms.pop(ref, (None, None))
         if comm:
             try:
@@ -336,47 +340,52 @@ class Syncable(Renderable):
             with edit_readonly(state):
                 state.busy = busy
 
-    def _process_bokeh_event(self, event: 'Event') -> None:
+    def _process_bokeh_event(self, doc: Document, event: Event) -> None:
         self._log('received bokeh event %s', event)
         busy = state.busy
         with edit_readonly(state):
             state.busy = True
         try:
-            self._process_event(event)
+            with set_curdoc(doc):
+                self._process_event(event)
         finally:
             self._log('finished processing bokeh event %s', event)
             with edit_readonly(state):
                 state.busy = busy
 
-    async def _change_coroutine(self, doc: Optional['Document'] = None) -> None:
+    async def _change_coroutine(self, doc: Document) -> None:
         if state._thread_pool:
-            state._thread_pool.submit(self._change_event, doc)
+            future = state._thread_pool.submit(self._change_event, doc)
+            future.add_done_callback(state._handle_future_exception)
         else:
             with set_curdoc(doc):
-                self._change_event(doc)
+                try:
+                    self._change_event(doc)
+                except Exception as e:
+                    state._handle_exception(e)
 
-    async def _event_coroutine(self, event, doc: 'Document') -> None:
+    async def _event_coroutine(self, doc: Document, event) -> None:
         if state._thread_pool:
-            state._thread_pool.submit(self._process_bokeh_event, event)
+            future = state._thread_pool.submit(self._process_bokeh_event, doc, event)
+            future.add_done_callback(state._handle_future_exception)
         else:
-            with set_curdoc(doc):
-                self._process_bokeh_event(event)
+            try:
+                self._process_bokeh_event(doc, event)
+            except Exception as e:
+                state._handle_exception(e)
 
-    def _change_event(self, doc: Optional['Document'] = None) -> None:
-        try:
-            state.curdoc = doc
-            events = self._events
-            self._events = {}
+    def _change_event(self, doc: Document) -> None:
+        events = self._events
+        self._events = {}
+        with set_curdoc(doc):
             self._process_events(events)
-        finally:
-            state.curdoc = None
 
-    def _schedule_change(self, doc: Optional['Document'], comm: Optional['Comm']) -> None:
+    def _schedule_change(self, doc: Document, comm: Comm | None) -> None:
         with hold(doc, comm=comm):
             self._change_event(doc)
 
     def _comm_change(
-        self, doc: 'Document', ref: str, comm: Optional['Comm'], subpath: str,
+        self, doc: Document, ref: str, comm: Comm | None, subpath: str,
         attr: str, old: Any, new: Any
     ) -> None:
         if subpath:
@@ -387,26 +396,39 @@ class Syncable(Renderable):
 
         self._events.update({attr: new})
         if state._thread_pool:
-            state._thread_pool.submit(self._schedule_change, doc, comm)
+            future = state._thread_pool.submit(self._schedule_change, doc, comm)
+            future.add_done_callback(state._handle_future_exception)
         else:
-            self._schedule_change(doc, comm)
+            try:
+                self._schedule_change(doc, comm)
+            except Exception as e:
+                state._handle_exception(e)
 
-    def _comm_event(self, event: 'Event') -> None:
+    def _comm_event(self, doc: Document, event: Event) -> None:
         if state._thread_pool:
-            state._thread_pool.submit(self._process_bokeh_event, event)
+            future = state._thread_pool.submit(self._process_bokeh_event, doc, event)
+            future.add_done_callback(state._handle_future_exception)
         else:
-            self._process_bokeh_event(event)
+            try:
+                self._process_bokeh_event(doc, event)
+            except Exception as e:
+                state._handle_exception(e)
 
-    def _server_event(self, doc: 'Document', event: 'Event') -> None:
+    def _register_events(self, *event_names: str, model: Model, doc: Document, comm: Comm | None) -> None:
+        for event_name in event_names:
+            method = self._comm_event if comm else self._server_event
+            model.on_event(event_name, partial(method, doc))
+
+    def _server_event(self, doc: Document, event: Event) -> None:
         if doc.session_context and not state._unblocked(doc):
             doc.add_next_tick_callback(
-                partial(self._event_coroutine, event, doc)
+                partial(self._event_coroutine, doc, event) # type: ignore
             )
         else:
-            self._comm_event(event)
+            self._comm_event(doc, event)
 
     def _server_change(
-        self, doc: 'Document', ref: str, subpath: str, attr: str,
+        self, doc: Document, ref: str, subpath: str, attr: str,
         old: Any, new: Any
     ) -> None:
         if subpath:
@@ -423,11 +445,14 @@ class Syncable(Renderable):
         if doc.session_context:
             cb = partial(self._change_coroutine, doc)
             if attr in self._priority_changes:
-                doc.add_next_tick_callback(cb)
+                doc.add_next_tick_callback(cb) # type: ignore
             else:
-                doc.add_timeout_callback(cb, self._debounce)
+                doc.add_timeout_callback(cb, self._debounce) # type: ignore
         else:
-            self._change_event(doc)
+            try:
+                self._change_event(doc)
+            except Exception as e:
+                state._handle_exception(e)
 
 
 class Reactive(Syncable, Viewable):
@@ -579,7 +604,7 @@ class Reactive(Syncable, Viewable):
             return controls.layout[0]
         return style.layout[0]
 
-    def jscallback(self, args: Mapping[str, Any]={}, **callbacks: str) -> 'Callback':
+    def jscallback(self, args: Dict[str, Any]={}, **callbacks: str) -> Callback:
         """
         Allows defining a JS callback to be triggered when a property
         changes on the source object. The keyword arguments define the
@@ -599,17 +624,13 @@ class Reactive(Syncable, Viewable):
         callback: Callback
           The Callback which can be used to disable the callback.
         """
-
         from .links import Callback
-        renamed = {
-            k: self._rename.get(v, v) for k, v in callbacks.items()
-        }
-        return Callback(self, code=renamed, args=args)
+        return Callback(self, code=callbacks, args=args)
 
     def jslink(
-        self, target: 'JSLinkTarget' , code: Dict[str, str] = None, args: Optional[Dict] = None,
+        self, target: JSLinkTarget , code: Dict[str, str] = None, args: Optional[Dict] = None,
         bidirectional: bool = False, **links: str
-    ) -> 'Link':
+    ) -> Link:
         """
         Links properties on the this Reactive object to those on the
         target Reactive object in JS code.
@@ -673,9 +694,9 @@ class SyncableData(Reactive):
         The currently selected rows in the data.""")
 
     # Parameters which when changed require an update of the data
-    _data_params: List[str] = []
+    _data_params: ClassVar[List[str]] = []
 
-    _rename: Mapping[str, str | None] = {'selection': None}
+    _rename: ClassVar[Mapping[str, str | None]] = {'selection': None}
 
     __abstract = True
 
@@ -730,8 +751,8 @@ class SyncableData(Reactive):
         self.param.update(**{self._data_params[0]: data})
 
     def _manual_update(
-        self, events: Tuple[param.parameterized.Event, ...], model: 'Model',
-        doc: 'Document', root: 'Model', parent: Optional['Model'], comm: 'Comm'
+        self, events: Tuple[param.parameterized.Event, ...], model: Model,
+        doc: Document, root: Model, parent: Optional[Model], comm: Comm
     ) -> None:
         for event in events:
             if event.type == 'triggered' and self._updating:
@@ -743,8 +764,9 @@ class SyncableData(Reactive):
     def _update_cds(self, *events: param.parameterized.Event) -> None:
         self._processed, self._data = self._get_data()
         msg = {'data': self._data}
+        named_events = {event.name: event for event in events}
         for ref, (m, _) in self._models.items():
-            self._apply_update(events, msg, m.source, ref)
+            self._apply_update(named_events, msg, m.source, ref)
 
     @updating
     def _update_selected(
@@ -752,10 +774,11 @@ class SyncableData(Reactive):
     ) -> None:
         indices = self.selection if indices is None else indices
         msg = {'indices': indices}
+        named_events = {event.name: event for event in events}
         for ref, (m, _) in self._models.items():
-            self._apply_update(events, msg, m.source.selected, ref)
+            self._apply_update(named_events, msg, m.source.selected, ref)
 
-    def _apply_stream(self, ref: str, model: 'Model', stream: 'DataDict', rollover: Optional[int]) -> None:
+    def _apply_stream(self, ref: str, model: Model, stream: 'DataDict', rollover: Optional[int]) -> None:
         self._changing[ref] = ['data']
         try:
             model.source.stream(stream, rollover)
@@ -778,7 +801,7 @@ class SyncableData(Reactive):
                 cb = partial(self._apply_stream, ref, m, stream, rollover)
                 doc.add_next_tick_callback(cb)
 
-    def _apply_patch(self, ref: str, model: 'Model', patch: 'Patches') -> None:
+    def _apply_patch(self, ref: str, model: Model, patch: 'Patches') -> None:
         self._changing[ref] = ['data']
         try:
             model.source.patch(patch)
@@ -978,7 +1001,7 @@ class SyncableData(Reactive):
         else:
             pd = None # type: ignore
         data = getattr(self, self._data_params[0])
-        patch_value_dict: 'Patches' = {}
+        patch_value_dict: Patches = {}
         if pd and isinstance(patch_value, pd.DataFrame):
             for column in patch_value.columns:
                 patch_value_dict[column] = []
@@ -1023,7 +1046,6 @@ class ReactiveData(SyncableData):
 
     def __init__(self, **params):
         super().__init__(**params)
-        self._old = None
 
     def _update_selection(self, indices: List[int]) -> None:
         self.selection = indices
@@ -1047,6 +1069,13 @@ class ReactiveData(SyncableData):
                         iv = dt.date.fromtimestamp(iv/1000)
                     new_values.append(iv)
                 converted = new_values
+        elif 'pandas' in sys.modules:
+            import pandas as pd
+            if Version(pd.__version__) >= Version('1.1.0'):
+                from pandas.core.arrays.masked import BaseMaskedDtype
+                if isinstance(dtype, BaseMaskedDtype):
+                    values = [dtype.na_value if v == '<NA>' else v for v in values]
+            converted = pd.Series(values).astype(dtype).values
         else:
             converted = values.astype(dtype)
         return values if converted is None else converted
@@ -1054,10 +1083,9 @@ class ReactiveData(SyncableData):
     def _process_data(self, data: Mapping[str, List | Dict[int, Any] | np.ndarray]) -> None:
         if self._updating:
             return
-
         # Get old data to compare to
         old_raw, old_data = self._get_data()
-        self._old = old_raw = old_raw.copy()
+        old_raw = old_raw.copy()
         if hasattr(old_raw, 'columns'):
             columns = list(old_raw.columns) # type: ignore
         else:
@@ -1123,11 +1151,11 @@ class ReactiveHTMLMetaclass(ParameterizedMetaclass):
     HTML attributes.
     """
 
-    _loaded_extensions: Set[str] = set()
+    _loaded_extensions: ClassVar[Set[str]] = set()
 
-    _name_counter: Counter = Counter()
+    _name_counter: ClassVar[Counter] = Counter()
 
-    _script_regex: str = r"script\([\"|'](.*)[\"|']\)"
+    _script_regex: ClassVar[str] = r"script\([\"|'](.*)[\"|']\)"
 
     def __init__(mcs, name: str, bases: Tuple[Type, ...], dict_: Mapping[str, Any]):
         from .io.datamodel import PARAM_MAPPING, construct_data_model
@@ -1349,24 +1377,26 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
                 `input` node in the example above.
     """
 
-    _child_config: Mapping[str, str] = {}
+    _child_config: ClassVar[Mapping[str, str]] = {}
 
-    _dom_events: Mapping[str, List[str]] = {}
+    _dom_events: ClassVar[Mapping[str, List[str]]] = {}
 
-    _extension_name: Optional[str] = None
+    _extension_name: ClassVar[Optional[str]] = None
 
-    _template: str = ""
+    _template: ClassVar[str] = ""
 
-    _scripts: Mapping[str, str | List[str]] = {}
+    _scripts: ClassVar[Mapping[str, str | List[str]]] = {}
 
-    _script_assignment: str = r'data\.([^[^\d\W]\w*)[ ]*[\+,\-,\*,\\,%,\*\*,<<,>>,>>>,&,\^,|,\&\&,\|\|,\?\?]*='
+    _script_assignment: ClassVar[str] = (
+        r'data\.([^[^\d\W]\w*)[ ]*[\+,\-,\*,\\,%,\*\*,<<,>>,>>>,&,\^,|,\&\&,\|\|,\?\?]*='
+    )
+
+    __css__: ClassVar[Optional[List[str]]] = None
+    __javascript__: ClassVar[Optional[List[str]]] = None
+    __javascript_modules__: ClassVar[Optional[List[str]]] = None
 
     __abstract = True
 
-    __css__: Optional[List[str]] = None
-    __javascript__: Optional[List[str]] = None
-    __javascript_modules__: Optional[List[str]] = None
-
     def __init__(self, **params):
         from .pane import panel
         for children_param in self._parser.children.values():
@@ -1405,7 +1435,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
             cls._extension_name in ReactiveHTMLMetaclass._loaded_extensions
         )
 
-    def _cleanup(self, root: 'Model' | None) -> None:
+    def _cleanup(self, root: Model | None = None) -> None:
         for child, panes in self._panes.items():
             for pane in panes:
                 pane._cleanup(root)
@@ -1420,9 +1450,9 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         return {}
 
     def _process_children(
-        self, doc: 'Document', root: 'Model', model: 'Model', comm: Optional['Comm'],
-        children: Mapping[str, str | 'Model' | List['Model']]
-    ) -> Mapping[str, str | 'Model' | List['Model']]:
+        self, doc: Document, root: Model, model: Model, comm: Optional[Comm],
+        children: Dict[str, List[Model]]
+    ) -> Dict[str, List[Model]]:
         return children
 
     def _init_params(self) -> Dict[str, Any]:
@@ -1479,11 +1509,11 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         return events
 
     def _get_children(
-        self, doc: 'Document', root: 'Model', model: 'Model', comm: Optional['Comm']
-    ) -> Mapping[str, str | 'Model' | List['Model']]:
+        self, doc: Document, root: Model, model: Model, comm: Optional[Comm]
+    ) -> Dict[str, List[Model]]:
         from .pane import panel
         old_models = model.children
-        new_models: Dict[str, List['Model']] = {parent: [] for parent in self._parser.children}
+        new_models: Dict[str, List[Model]] = {parent: [] for parent in self._parser.children}
         new_panes: Dict[str, List[Viewable] | Dict[str, Viewable] | None] = {}
         internal_panes: Dict[str, List[Viewable] | None] = {}
 
@@ -1551,8 +1581,9 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         #  {% endfor %}
         template_string = self._template
         for var, obj in self._parser.loop_map.items():
-            template_string = template_string.replace(
-                '${%s}' % var, '${%s[{{ loop.index0 }}]}' % obj)
+            for var in self._parser.loop_var_map[var]:
+                template_string = template_string.replace(
+                    '${%s}' % var, '${%s[{{ loop.index0 }}]}' % obj)
 
         # Add index to templated loop node ids
         for dom_node, _ in self._parser.looped:
@@ -1628,9 +1659,9 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         return linked_properties
 
     def _get_model(
-        self, doc: 'Document', root: Optional['Model'] = None,
-        parent: Optional['Model'] = None, comm: Optional['Comm'] = None
-    ) -> 'Model':
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         properties = self._process_param_change(self._init_params())
         model = _BkReactiveHTML(**properties)
         if comm and not self._loaded():
@@ -1653,17 +1684,13 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         if not root:
             root = model
 
-        for p, v in model.data.properties_with_values().items():
+        data_model: DataModel = model.data # type: ignore
+        for p, v in data_model.properties_with_values().items():
             if isinstance(v, DataModel):
                 v.tags.append(f"__ref:{root.ref['id']}")
-        model.children = self._get_children(doc, root, model, comm)
-
-        if comm:
-            model.on_event('dom_event', self._comm_event)
-        else:
-            model.on_event('dom_event', partial(self._server_event, doc))
-
-        self._link_props(model.data, self._linked_properties(), doc, root, comm)
+        model.update(children=self._get_children(doc, root, model, comm))
+        self._register_events('dom_event', model=model, doc=doc, comm=comm)
+        self._link_props(data_model, self._linked_properties(), doc, root, comm)
         self._models[root.ref['id']] = (model, parent)
         return model
 
@@ -1676,8 +1703,12 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         event_type = event.data['type']
         star_cbs = self._event_callbacks.get('*', {})
         node_cbs = self._event_callbacks.get(event.node, {})
+
+        def match(node, pattern):
+            return re.findall(re.sub(r'\{\{.*loop.index.*\}\}', r'\\d+', pattern), node)
+
         inline_cbs = {attr: [getattr(self, p)] for node, attr, p in self._inline_callbacks
-                      if node == event.node}
+                      if node == event.node or match(event.node, node)}
         event_cbs = (
             node_cbs.get(event_type, []) + node_cbs.get('*', []) +
             star_cbs.get(event_type, []) + star_cbs.get('*', []) +
@@ -1686,7 +1717,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         for cb in event_cbs:
             cb(event)
 
-    def _set_on_model(self, msg: Mapping[str, Any], root: 'Model', model: 'Model') -> None:
+    def _set_on_model(self, msg: Mapping[str, Any], root: Model, model: Model) -> None:
         if not msg:
             return
         old = self._changing.get(root.ref['id'], [])
@@ -1704,7 +1735,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
 
     def _update_model(
         self, events: Dict[str, param.parameterized.Event], msg: Dict[str, Any],
-        root: 'Model', model: 'Model', doc: 'Document', comm: Optional['Comm']
+        root: Model, model: Model, doc: Document, comm: Optional[Comm]
     ) -> None:
         child_params = self._parser.children.values()
         new_children, model_msg, data_msg  = {}, {}, {}
@@ -1765,7 +1796,7 @@ class ReactiveHTML(Reactive, metaclass=ReactiveHTMLMetaclass):
         self._event_callbacks[node][event].append(callback)
         events = self._get_events()
         for ref, (model, _) in self._models.items():
-            self._apply_update([], {'events': events}, model, ref)
+            self._apply_update({}, {'events': events}, model, ref)
 
 __all__ = (
     "Reactive",
