@@ -1,10 +1,11 @@
-import os
-import json
 import glob
-import pytest
+import json
+import os
 
 from io import StringIO
 
+import pytest
+
 from bokeh.models import CustomJS
 
 from panel import Row
@@ -12,7 +13,9 @@ from panel.config import config
 from panel.io.embed import embed_state
 from panel.pane import Str
 from panel.param import Param
-from panel.widgets import IntSlider, Select, FloatSlider, Checkbox, StaticText
+from panel.widgets import (
+    Checkbox, FloatSlider, IntSlider, Select, StaticText,
+)
 
 
 def test_embed_param_jslink(document, comm):
@@ -534,3 +537,13 @@ def test_save_embed_json(tmpdir):
         assert event['kind'] == 'ModelChanged'
         assert event['attr'] == 'text'
         assert event['new'] == '&lt;pre&gt;%s&lt;/pre&gt;' % v
+
+def test_embed_widget_disabled(document, comm):
+    select = Select(options=['A', 'B', 'C'], disabled=True)
+    string = Str()
+    select.link(string, value='object')
+    string.param.watch(print, 'object')
+    panel = Row(select, string)
+    with config.set(embed=True):
+        model = panel.get_root(document, comm)
+    assert embed_state(panel, model, document) is None
