@@ -1,7 +1,11 @@
-import panel as pn
 import pytest
+
+import panel as pn
+
 from panel.widgets import Grammar, GrammarList, SpeechToText
-from panel.widgets.speech_to_text import RecognitionAlternative, RecognitionResult
+from panel.widgets.speech_to_text import (
+    RecognitionAlternative, RecognitionResult,
+)
 
 CLASSES = [
     Grammar,
@@ -272,6 +276,6 @@ def test_get_color_app():
 
 
 if __name__.startswith("bokeh"):
-    pn.config.sizing_mode = "stretch_width"
+    pn.extension(sizing_mode="stretch_width")
     # test_get_advanced_app().servable()
     test_get_color_app().servable()
