@@ -1,24 +1,25 @@
 import datetime as dt
 import time
 
-from packaging.version import Version
-
 import numpy as np
 import pytest
 import requests
 
+from packaging.version import Version
+
 try:
     import pandas as pd
+
     from pandas._testing import (
-        makeCustomDataframe, makeMixedDataFrame, makeTimeDataFrame
+        makeCustomDataframe, makeMixedDataFrame, makeTimeDataFrame,
     )
 except ImportError:
     pytestmark = pytest.mark.skip('pandas not available')
 
 from bokeh.models.widgets.tables import (
-    NumberFormatter, IntEditor, NumberEditor, StringFormatter,
-    SelectEditor, DateFormatter, DataCube, CellEditor,
-    SumAggregator, AvgAggregator, MinAggregator
+    AvgAggregator, CellEditor, CheckboxEditor, DataCube, DateEditor,
+    DateFormatter, IntEditor, MinAggregator, NumberEditor, NumberFormatter,
+    SelectEditor, StringEditor, StringFormatter, SumAggregator,
 )
 
 from panel.depends import bind
@@ -251,7 +252,7 @@ def test_tabulator_multi_index(document, comm):
         {'field': 'A'},
         {'field': 'C'},
         {'field': 'B'},
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
 
     assert np.array_equal(model.source.data['A'], np.array([0., 1., 2., 3., 4.]))
@@ -268,7 +269,7 @@ def test_tabulator_multi_index_remote_pagination(document, comm):
         {'field': 'A'},
         {'field': 'C'},
         {'field': 'B'},
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
 
     assert np.array_equal(model.source.data['A'], np.array([0., 1., 2.]))
@@ -385,7 +386,7 @@ def test_tabulator_config_defaults(document, comm):
         {'field': 'A'},
         {'field': 'B'},
         {'field': 'C'},
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
     assert model.configuration['selectable'] == True
 
@@ -400,7 +401,7 @@ def test_tabulator_config_widths_percent(document, comm):
         {'field': 'A', 'width': '22%'},
         {'field': 'B'},
         {'field': 'C'},
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
     assert model.columns[2].width == 100
 
@@ -415,12 +416,12 @@ def test_tabulator_header_filters_config_boolean(document, comm):
         {'field': 'A', 'headerFilter': True},
         {'field': 'B', 'headerFilter': True},
         {'field': 'C', 'headerFilter': True},
-        {'field': 'D', 'headerFilter': True}
+        {'field': 'D', 'headerFilter': False, 'sorter': 'timestamp'} # Datetime header filtering not supported
     ]
 
-def test_tabulator_header_filters_column_config_select(document, comm):
+def test_tabulator_header_filters_column_config_list(document, comm):
     df = makeMixedDataFrame()
-    table = Tabulator(df, header_filters={'C': 'select'})
+    table = Tabulator(df, header_filters={'C': 'list'})
 
     model = table.get_root(document, comm)
 
@@ -428,15 +429,34 @@ def test_tabulator_header_filters_column_config_select(document, comm):
         {'field': 'index'},
         {'field': 'A'},
         {'field': 'B'},
-        {'field': 'C', 'headerFilter': 'select', 'headerFilterParams': {'values': True}},
-        {'field': 'D'}
+        {'field': 'C', 'headerFilter': 'list', 'headerFilterParams': {'valuesLookup': True}},
+        {'field': 'D', 'sorter': 'timestamp'}
+    ]
+    assert model.configuration['selectable'] == True
+
+@pytest.mark.parametrize('editor', ['select', 'autocomplete'])
+def test_tabulator_header_filters_column_config_select_autocomplete_backwards_compat(document, comm, editor):
+    df = makeMixedDataFrame()
+    table = Tabulator(df, header_filters={
+        'C': editor,
+        'D': {'type': editor, 'values': True}
+    })
+
+    model = table.get_root(document, comm)
+
+    assert model.configuration['columns'] == [
+        {'field': 'index'},
+        {'field': 'A'},
+        {'field': 'B'},
+        {'field': 'C', 'headerFilter': 'list', 'headerFilterParams': {'valuesLookup': True}},
+        {'field': 'D', 'headerFilter': 'list', 'headerFilterParams': {'valuesLookup': True}, 'sorter': 'timestamp'},
     ]
     assert model.configuration['selectable'] == True
 
 def test_tabulator_header_filters_column_config_dict(document, comm):
     df = makeMixedDataFrame()
     table = Tabulator(df, header_filters={
-        'C': {'type': 'select', 'values': True, 'func': '!=', 'placeholder': 'Not equal'}
+        'C': {'type': 'list', 'valuesLookup': True, 'func': '!=', 'placeholder': 'Not equal'}
     })
 
     model = table.get_root(document, comm)
@@ -447,15 +467,60 @@ def test_tabulator_header_filters_column_config_dict(document, comm):
         {'field': 'B'},
         {
             'field': 'C',
-            'headerFilter': 'select',
-            'headerFilterParams': {'values': True},
+            'headerFilter': 'list',
+            'headerFilterParams': {'valuesLookup': True},
             'headerFilterFunc': '!=',
             'headerFilterPlaceholder': 'Not equal'
         },
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
     assert model.configuration['selectable'] == True
 
+
+def test_tabulator_editors_default(document, comm):
+    df = pd.DataFrame({
+        'int': [1, 2],
+        'float': [3.14, 6.28],
+        'str': ['A', 'B'],
+        'date': [dt.date(2009, 1, 8), dt.date(2010, 1, 8)],
+        'datetime': [dt.datetime(2009, 1, 8), dt.datetime(2010, 1, 8)],
+        'bool': [True, False],
+    })
+    table = Tabulator(df)
+    model = table.get_root(document, comm)
+    assert isinstance(model.columns[1].editor, IntEditor)
+    assert isinstance(model.columns[2].editor, NumberEditor)
+    assert isinstance(model.columns[3].editor, StringEditor)
+    assert isinstance(model.columns[4].editor, DateEditor)
+    assert isinstance(model.columns[5].editor, DateEditor)
+    assert isinstance(model.columns[6].editor, CheckboxEditor)
+
+
+def test_tabulator_formatters_default(document, comm):
+    df = pd.DataFrame({
+        'int': [1, 2],
+        'float': [3.14, 6.28],
+        'str': ['A', 'B'],
+        'date': [dt.date(2009, 1, 8), dt.date(2010, 1, 8)],
+        'datetime': [dt.datetime(2009, 1, 8), dt.datetime(2010, 1, 8)],
+    })
+    table = Tabulator(df)
+    model = table.get_root(document, comm)
+    mformatter = model.columns[1].formatter
+    assert isinstance(mformatter, NumberFormatter)
+    mformatter = model.columns[2].formatter
+    assert isinstance(mformatter, NumberFormatter)
+    assert mformatter.format == '0,0.0[00000]'
+    mformatter = model.columns[3].formatter
+    assert isinstance(mformatter, StringFormatter)
+    mformatter = model.columns[4].formatter
+    assert isinstance(mformatter, DateFormatter)
+    assert mformatter.format == '%Y-%m-%d'
+    mformatter = model.columns[5].formatter
+    assert isinstance(mformatter, DateFormatter)
+    assert mformatter.format == '%Y-%m-%d %H:%M:%S'
+
+
 def test_tabulator_config_formatter_string(document, comm):
     df = makeMixedDataFrame()
     table = Tabulator(df, formatters={'B': 'tickCross'})
@@ -474,22 +539,31 @@ def test_tabulator_config_formatter_dict(document, comm):
     assert model.configuration['columns'][2] == {'field': 'B', 'formatter': 'tickCross', 'formatterParams': {'tristate': True}}
 
 
-def test_tabulator_config_editor_string(document, comm):
+def test_tabulator_config_editor_string_backwards_compat(document, comm):
     df = makeMixedDataFrame()
     table = Tabulator(df, editors={'B': 'select'})
 
     model = table.get_root(document, comm)
 
-    assert model.configuration['columns'][2] == {'field': 'B', 'editor': 'select'}
+    assert model.configuration['columns'][2] == {'field': 'B', 'editor': 'list'}
+
+
+def test_tabulator_config_editor_string(document, comm):
+    df = makeMixedDataFrame()
+    table = Tabulator(df, editors={'B': 'list'})
+
+    model = table.get_root(document, comm)
+
+    assert model.configuration['columns'][2] == {'field': 'B', 'editor': 'list'}
 
 
 def test_tabulator_config_editor_dict(document, comm):
     df = makeMixedDataFrame()
-    table = Tabulator(df, editors={'B': {'type': 'select', 'values': True}})
+    table = Tabulator(df, editors={'B': {'type': 'list', 'valuesLookup': True}})
 
     model = table.get_root(document, comm)
 
-    assert model.configuration['columns'][2] == {'field': 'B', 'editor': 'select', 'editorParams': {'values': True}}
+    assert model.configuration['columns'][2] == {'field': 'B', 'editor': 'list', 'editorParams': {'valuesLookup': True}}
 
 
 def test_tabulator_groups(document, comm):
@@ -508,7 +582,7 @@ def test_tabulator_groups(document, comm):
         {'title': 'Other',
          'columns': [
             {'field': 'C'},
-            {'field': 'D'}
+            {'field': 'D', 'sorter': 'timestamp'}
         ]}
     ]
 
@@ -541,7 +615,7 @@ def test_tabulator_frozen_cols(document, comm):
         {'field': 'A'},
         {'field': 'B'},
         {'field': 'C'},
-        {'field': 'D'}
+        {'field': 'D', 'sorter': 'timestamp'}
     ]
 
 
@@ -567,6 +641,20 @@ def test_tabulator_selectable_rows(document, comm):
     assert model.selectable_rows == [3, 4]
 
 
+@pytest.mark.xfail(reason='See https://github.com/holoviz/panel/issues/3644')
+def test_tabulator_selectable_rows_nonallowed_selection_error(document, comm):
+    df = makeMixedDataFrame()
+    table = Tabulator(df, selectable_rows=lambda df: [1])
+
+    model = table.get_root(document, comm)
+
+    assert model.selectable_rows == [1]
+
+    #
+    with pytest.raises(ValueError):
+        table.selection = [0]
+
+
 def test_tabulator_pagination(document, comm):
     df = makeMixedDataFrame()
     table = Tabulator(df, pagination='remote', page_size=2)
@@ -1499,18 +1587,25 @@ def test_tabulator_widget_scalar_filter(document, comm):
     for col, values in model.source.data.items():
         np.testing.assert_array_equal(values, expected[col])
 
-def test_tabulator_constant_list_filter(document, comm):
+@pytest.mark.parametrize('col', ['A', 'B', 'C', 'D'])
+def test_tabulator_constant_list_filter(document, comm, col):
     df = makeMixedDataFrame()
+    # The mixed dataframe has duplicate number values in the B columns,
+    # simplify the test by setting the targetted valued before filtering.
+    df.at[2, 'B'] = 10.0
+    df.at[4, 'B'] = 20.0
     table = Tabulator(df)
 
     model = table.get_root(document, comm)
 
-    table.add_filter(['foo3', 'foo5'], 'C')
+    values = list(df.iloc[[2, 4], :][col])
+
+    table.add_filter(values, col)
 
     expected = {
         'index': np.array([2, 4]),
         'A': np.array([2., 4.]),
-        'B': np.array([0., 0.]),
+        'B': np.array([10., 20.]),
         'C': np.array(['foo3', 'foo5']),
         'D': np.array(['2009-01-05T00:00:00.000000000',
                        '2009-01-07T00:00:00.000000000'],
@@ -1694,17 +1789,15 @@ def test_tabulator_patch_event():
             table._process_event(event)
             assert values[-1] == (col, row, df[col].iloc[row])
 
-
-
-def test_server_edit_event():
+def test_server_edit_event(port):
     df = makeMixedDataFrame()
     table = Tabulator(df)
 
-    serve(table, port=7001, threaded=True, show=False)
+    serve(table, port=port, threaded=True, show=False)
 
     time.sleep(0.5)
 
-    requests.get('http://localhost:7001')
+    requests.get(f'http://localhost:{port}')
 
     assert table._models
     ref, (model, _) = list(table._models.items())[0]
@@ -1738,7 +1831,7 @@ def test_tabulator_cell_click_event():
             table._process_event(event)
             assert values[-1] == (col, row, data[col].iloc[row])
 
-def test_server_cell_click_async_event():
+def test_server_cell_click_async_event(port):
     df = makeMixedDataFrame()
     table = Tabulator(df)
 
@@ -1752,7 +1845,6 @@ def test_server_cell_click_async_event():
 
     table.on_click(cb)
 
-    port = 7002
     serve(table, port=port, threaded=True, show=False)
 
     # Wait for server to start
@@ -1789,3 +1881,14 @@ def test_tabulator_pagination_remote_cell_click_event():
                 event = CellClickEvent(model=None, column=col, row=row)
                 table._process_event(event)
                 assert values[-1] == (col, (p*2)+row, data[col].iloc[(p*2)+row])
+
+def test_tabulator_cell_click_event_error_duplicate_index():
+    df = pd.DataFrame(data={'A': [1, 2]}, index=['a', 'a'])
+    table = Tabulator(df, sorters=[{'field': 'A', 'dir': 'desc'}])
+
+    values = []
+    table.on_click(lambda e: values.append((e.column, e.row, e.value)))
+
+    event = CellClickEvent(model=None, column='y', row=0)
+    with pytest.raises(ValueError, match="Found this duplicate index: 'a'"):
+        table._process_event(event)
