@@ -22,13 +22,14 @@ from functools import partial
 from html import escape  # noqa
 from importlib import import_module
 from typing import (
-    Any, AnyStr, Dict, Iterable, Iterator, List, Optional, Union
+    Any, AnyStr, Dict, Iterable, Iterator, List, Optional, Union,
 )
 
 import bokeh
 import numpy as np
 import param
 
+from bokeh.model import Model
 from packaging.version import Version
 
 datetime_types = (np.datetime64, dt.datetime, dt.date)
@@ -36,6 +37,9 @@ datetime_types = (np.datetime64, dt.datetime, dt.date)
 bokeh_version = Version(bokeh.__version__)
 
 
+PARAM_NAME_PATTERN = re.compile(r'^.*\d{5}$')
+
+
 def isfile(path: str) -> bool:
     """Safe version of os.path.isfile robust to path length issues on Windows"""
     try:
@@ -222,6 +226,15 @@ def is_parameterized(obj) -> bool:
     return (isinstance(obj, param.Parameterized) or
             (isinstance(obj, type) and issubclass(obj, param.Parameterized)))
 
+def is_holoviews(obj: Any) -> bool:
+    """
+    Whether the object is a HoloViews type that can be rendered.
+    """
+    if 'holoviews' not in sys.modules:
+        return False
+    from holoviews.core.dimension import Dimensioned
+    from holoviews.plotting import Plot
+    return isinstance(obj, (Dimensioned, Plot))
 
 def isdatetime(value) -> bool:
     """
@@ -360,7 +373,11 @@ def edit_readonly(parameterized: param.Parameterized) -> Iterator:
 
 def lazy_load(module, model, notebook=False, root=None):
     if module in sys.modules:
-        return getattr(sys.modules[module], model)
+        model_cls = getattr(sys.modules[module], model)
+        if model not in Model.model_class_reverse_map:
+            Model.model_class_reverse_map[model] = model_cls
+        return model_cls
+
     if notebook:
         ext = module.split('.')[-1]
         param.main.param.warning(
@@ -429,3 +446,25 @@ def fullpath(path: Union[AnyStr, os.PathLike]) -> Union[AnyStr, os.PathLike]:
     """Expanduser and then abspath for a given path
     """
     return os.path.abspath(os.path.expanduser(path))
+
+
+def base_version(version: str) -> str:
+    """Extract the final release and if available pre-release (alpha, beta,
+    release candidate) segments of a PEP440 version, defined with three
+    components (major.minor.micro).
+
+    Useful to avoid nbsite/sphinx to display the documentation HTML title
+    with a not so informative and rather ugly long version (e.g.
+    ``0.13.0a19.post4+g0695e214``). Use it in ``conf.py``::
+
+        version = release = base_version(package.__version__)
+
+    Return the version passed as input if no match is found with the pattern.
+    """
+    # look at the start for e.g. 0.13.0, 0.13.0rc1, 0.13.0a19, 0.13.0b10
+    pattern = r"([\d]+\.[\d]+\.[\d]+(?:a|rc|b)?[\d]*)"
+    match = re.match(pattern, version)
+    if match:
+        return match.group()
+    else:
+        return version
