@@ -1,42 +1,59 @@
 """
 Defines Player widgets which offer media-player like controls.
 """
+from __future__ import annotations
+
+from typing import (
+    TYPE_CHECKING, ClassVar, Mapping, Type,
+)
+
 import param
 
+from ..config import config
 from ..models.widgets import Player as _BkPlayer
-from ..util import isIn, indexOf
+from ..util import indexOf, isIn
 from .base import Widget
 from .select import SelectBase
 
+if TYPE_CHECKING:
+    from bokeh.model import Model
+
 
 class PlayerBase(Widget):
 
-    interval = param.Integer(default=500, doc="Interval between updates")
-
-    loop_policy = param.ObjectSelector(default='once',
-                                       objects=['once', 'loop', 'reflect'], doc="""
-        Policy used when player hits last frame""")
-
-    step = param.Integer(default=1, doc="""
-        Number of frames to step forward and back by on each event.""")
-
-    show_loop_controls = param.Boolean(default=True, doc="""
-        Whether the loop controls radio buttons are shown""")
-
     direction = param.Integer(0, doc="""
         Current play direction of the Player (-1: playing in reverse,
         0: paused, 1: playing)""")
 
+    interval = param.Integer(default=500, doc="""
+        Interval between updates, in milliseconds. Default is 500, i.e.
+        two updates per second.""")
+
+    loop_policy = param.ObjectSelector(
+        default='once', objects=['once', 'loop', 'reflect'], doc="""
+        Policy used when player hits last frame""")
+
+    show_loop_controls = param.Boolean(default=True, doc="""
+        Whether the loop controls radio buttons are shown""")
+
+    step = param.Integer(default=1, doc="""
+        Number of frames to step forward and back by on each event.""")
+
     height = param.Integer(default=80)
 
     width = param.Integer(default=510)
 
-    _widget_type = _BkPlayer
+    _rename: ClassVar[Mapping[str, str | None]] = {'name': None}
 
-    _rename = {'name': None}
+    _widget_type: ClassVar[Type[Model]] = _BkPlayer
 
     __abstract = True
 
+    def __init__(self, **params):
+        if 'value' in params and 'value_throttled' in self.param:
+            params['value_throttled'] = params['value']
+        super().__init__(**params)
+
     def play(self):
         self.direction = 1
 
@@ -52,7 +69,7 @@ class Player(PlayerBase):
     """
     The `Player` provides controls to play and skip through a number of
     frames defined by explicit start and end values.  The speed at
-    which the widget plays is defined by the `interval`, but it is also
+    which the widget plays is defined by the `interval` (in milliseconds), but it is also
     possible to skip frames using the `step` parameter.
 
     Reference: https://panel.holoviz.org/reference/widgets/Player.html
@@ -68,7 +85,10 @@ class Player(PlayerBase):
 
     value = param.Integer(default=0, doc="Current player value")
 
-    _supports_embed = True
+    value_throttled = param.Integer(default=0, constant=True, doc="""
+        Current throttled player value.""")
+
+    _supports_embed: ClassVar[bool] = True
 
     def __init__(self, **params):
         if 'length' in params:
@@ -80,6 +100,14 @@ class Player(PlayerBase):
             params['value'] = params['start']
         super().__init__(**params)
 
+    def _process_property_change(self, msg):
+        if config.throttled:
+            if "value" in msg:
+                del msg["value"]
+            if "value_throttled" in msg:
+                msg["value"] = msg["value_throttled"]
+        return super()._process_property_change(msg)
+
     def _get_embed_state(self, root, values=None, max_opts=3):
         if values is None:
             values = list(range(self.start, self.end, self.step))
@@ -92,7 +120,7 @@ class DiscretePlayer(PlayerBase, SelectBase):
     """
     The `DiscretePlayer` provides controls to iterate through a list of
     discrete options.  The speed at which the widget plays is defined
-    by the `interval`, but it is also possible to skip items using the
+    by the `interval` (in milliseconds), but it is also possible to skip items using the
     `step` parameter.
 
     Reference: https://panel.holoviz.org/reference/widgets/DiscretePlayer.html
@@ -110,9 +138,11 @@ class DiscretePlayer(PlayerBase, SelectBase):
 
     value = param.Parameter(doc="Current player value")
 
-    _rename = {'name': None, 'options': None}
+    value_throttled = param.Parameter(constant=True, doc="Current player value")
 
-    _source_transforms = {'value': None}
+    _rename: ClassVar[Mapping[str, str | None]] = {'name': None, 'options': None}
+
+    _source_transforms: ClassVar[Mapping[str, str | None]] = {'value': None, 'value_throttled': None}
 
     def _process_param_change(self, msg):
         values = self.values
@@ -130,8 +160,9 @@ class DiscretePlayer(PlayerBase, SelectBase):
         return super()._process_param_change(msg)
 
     def _process_property_change(self, msg):
-        if 'value' in msg:
-            value = msg.pop('value')
-            if value < len(self.options):
-                msg['value'] = self.values[value]
+        for prop in ('value', 'value_throttled'):
+            if prop in msg:
+                value = msg.pop(prop)
+                if value < len(self.options):
+                    msg[prop] = self.values[value]
         return msg
