@@ -2,26 +2,35 @@
 Defines various Select widgets which allow choosing one or more items
 from a list of options.
 """
+from __future__ import annotations
+
+import itertools
 import re
 
 from collections import OrderedDict
-import itertools
+from typing import (
+    TYPE_CHECKING, ClassVar, Mapping, Type,
+)
 
 import param
 
 from bokeh.models.widgets import (
-    AutocompleteInput as _BkAutocompleteInput, CheckboxGroup as _BkCheckboxGroup,
-    CheckboxButtonGroup as _BkCheckboxButtonGroup, MultiSelect as _BkMultiSelect,
-    RadioButtonGroup as _BkRadioButtonGroup, RadioGroup as _BkRadioBoxGroup,
-    MultiChoice as _BkMultiChoice
+    AutocompleteInput as _BkAutocompleteInput,
+    CheckboxButtonGroup as _BkCheckboxButtonGroup,
+    CheckboxGroup as _BkCheckboxGroup, MultiChoice as _BkMultiChoice,
+    MultiSelect as _BkMultiSelect, RadioButtonGroup as _BkRadioButtonGroup,
+    RadioGroup as _BkRadioBoxGroup,
 )
 
 from ..layout import Column, VSpacer
-from ..models import SingleSelect as _BkSingleSelect, CustomSelect
-from ..util import isIn, indexOf
-from .base import Widget, CompositeWidget
-from .button import _ButtonBase, Button
-from .input import TextInput, TextAreaInput
+from ..models import CustomSelect, SingleSelect as _BkSingleSelect
+from ..util import PARAM_NAME_PATTERN, indexOf, isIn
+from .base import CompositeWidget, Widget
+from .button import Button, _ButtonBase
+from .input import TextAreaInput, TextInput
+
+if TYPE_CHECKING:
+    from bokeh.model import Model
 
 
 class SelectBase(Widget):
@@ -32,7 +41,13 @@ class SelectBase(Widget):
 
     @property
     def labels(self):
-        return [str(o) for o in self.options]
+        labels = []
+        for o in self.options:
+            if isinstance(o, param.Parameterized) and not PARAM_NAME_PATTERN.match(o.name):
+                labels.append(o.name)
+            else:
+                labels.append(str(o))
+        return labels
 
     @property
     def values(self):
@@ -51,7 +66,7 @@ class SingleSelectBase(SelectBase):
 
     value = param.Parameter(default=None)
 
-    _supports_embed = True
+    _supports_embed: ClassVar[bool] = True
 
     __abstract = True
 
@@ -156,7 +171,9 @@ class Select(SingleSelectBase):
         If set to 1 displays options as dropdown otherwise displays
         scrollable area.""")
 
-    _source_transforms = {'size': None, 'groups': None}
+    _source_transforms: ClassVar[Mapping[str, str | None]] = {
+        'size': None, 'groups': None
+    }
 
     @property
     def _widget_type(self):
@@ -291,7 +308,7 @@ class _MultiSelectBase(SingleSelectBase):
 
     value = param.List(default=[])
 
-    _supports_embed = False
+    _supports_embed: ClassVar[bool] = False
 
     def _process_param_change(self, msg):
         msg = super(SingleSelectBase, self)._process_param_change(msg)
@@ -339,7 +356,7 @@ class MultiSelect(_MultiSelectBase):
         The number of items displayed at once (i.e. determines the
         widget height).""")
 
-    _widget_type = _BkMultiSelect
+    _widget_type: ClassVar[Type[Model]] = _BkMultiSelect
 
 
 class MultiChoice(_MultiSelectBase):
@@ -383,10 +400,7 @@ class MultiChoice(_MultiSelectBase):
     solid = param.Boolean(default=True, doc="""
         Whether to display widget with solid or light style.""")
 
-    _widget_type = _BkMultiChoice
-
-
-_AutocompleteInput_rename = {'name': 'title', 'options': 'completions'}
+    _widget_type: ClassVar[Type[Model]] = _BkMultiChoice
 
 
 class AutocompleteInput(Widget):
@@ -435,9 +449,9 @@ class AutocompleteInput(Widget):
     value_input = param.String(default='', allow_None=True, doc="""
       Initial or entered text value updated on every key press.""")
 
-    _widget_type = _BkAutocompleteInput
+    _rename: ClassVar[Mapping[str, str | None]] = {'name': 'title', 'options': 'completions'}
 
-    _rename = _AutocompleteInput_rename
+    _widget_type: ClassVar[Type[Model]] = _BkAutocompleteInput
 
     def _process_param_change(self, msg):
         msg = super()._process_param_change(msg)
@@ -451,7 +465,7 @@ class _RadioGroupBase(SingleSelectBase):
 
     _supports_embed = False
 
-    _rename = {'name': None, 'options': 'labels', 'value': 'active'}
+    _rename: ClassVar[Mapping[str, str | None]] = {'name': None, 'options': 'labels', 'value': 'active'}
 
     _source_transforms = {'value': "source.labels[value]"}
 
@@ -523,9 +537,9 @@ class RadioButtonGroup(_RadioGroupBase, _ButtonBase):
         objects=['horizontal', 'vertical'], doc="""
         Button group orientation, either 'horizontal' (default) or 'vertical'.""")
 
-    _widget_type = _BkRadioButtonGroup
+    _supports_embed: ClassVar[bool] = True
 
-    _supports_embed = True
+    _widget_type: ClassVar[Type[Model]] = _BkRadioButtonGroup
 
 
 
@@ -551,9 +565,9 @@ class RadioBoxGroup(_RadioGroupBase):
         Whether the items be arrange vertically (``False``) or
         horizontally in-line (``True``).""")
 
-    _supports_embed = True
+    _supports_embed: ClassVar[bool] = True
 
-    _widget_type = _BkRadioBoxGroup
+    _widget_type: ClassVar[Type[Model]] = _BkRadioBoxGroup
 
 
 
@@ -561,7 +575,7 @@ class _CheckGroupBase(SingleSelectBase):
 
     value = param.List(default=[])
 
-    _rename = {'name': None, 'options': 'labels', 'value': 'active'}
+    _rename: ClassVar[Mapping[str, str | None]] = {'name': None, 'options': 'labels', 'value': 'active'}
 
     _source_transforms = {'value': "value.map((index) => source.labels[index])"}
 
@@ -618,7 +632,7 @@ class CheckButtonGroup(_CheckGroupBase, _ButtonBase):
         objects=['horizontal', 'vertical'], doc="""
         Button group orientation, either 'horizontal' (default) or 'vertical'.""")
 
-    _widget_type = _BkCheckboxButtonGroup
+    _widget_type: ClassVar[Type[Model]] = _BkCheckboxButtonGroup
 
 
 class CheckBoxGroup(_CheckGroupBase):
@@ -644,7 +658,7 @@ class CheckBoxGroup(_CheckGroupBase):
         Whether the items be arrange vertically (``False``) or
         horizontally in-line (``True``).""")
 
-    _widget_type = _BkCheckboxGroup
+    _widget_type: ClassVar[Type[Model]] = _BkCheckboxGroup
 
 
 
