@@ -1,19 +1,24 @@
+from __future__ import annotations
+
 import datetime as dt
 import uuid
 
 from functools import partial
 from types import FunctionType, MethodType
+from typing import (
+    TYPE_CHECKING, Any, Callable, ClassVar, Dict, List, Mapping, Optional,
+    Tuple, Type, TypeVar,
+)
 
 import numpy as np
 import param
 
 from bokeh.models import ColumnDataSource
 from bokeh.models.widgets.tables import (
-    AvgAggregator, CellEditor, CellFormatter, CheckboxEditor,
-    DataCube, DataTable, DateEditor, DateFormatter, GroupingInfo,
-    IntEditor, MaxAggregator, MinAggregator, NumberEditor,
-    NumberFormatter, RowAggregator, StringEditor, StringFormatter,
-    SumAggregator, TableColumn
+    AvgAggregator, CellEditor, CellFormatter, CheckboxEditor, DataCube,
+    DataTable, DateEditor, DateFormatter, GroupingInfo, IntEditor,
+    MaxAggregator, MinAggregator, NumberEditor, NumberFormatter, RowAggregator,
+    StringEditor, StringFormatter, SumAggregator, TableColumn,
 )
 from bokeh.util.serialization import convert_datetime_array
 from pyviz_comms import JupyterComm
@@ -22,12 +27,28 @@ from ..depends import param_value_if_widget
 from ..io.resources import LOCAL_DIST, set_resource_mode
 from ..io.state import state
 from ..reactive import ReactiveData
+from ..util import (
+    clone_model, isdatetime, lazy_load, updating,
+)
 from ..viewable import Layoutable
-from ..util import clone_model, isdatetime, lazy_load, updating
 from .base import Widget
 from .button import Button
 from .input import TextInput
 
+if TYPE_CHECKING:
+    try:
+        import pandas as pd
+        DataFrameType = pd.DataFrame
+    except Exception:
+        DataFrameType = TypeVar('DataFrameType')
+
+    from bokeh.document import Document
+    from bokeh.model import Model
+    from bokeh.models.sources import DataDict
+    from pyviz_comms import Comm
+
+    from ..models.tabulator import CellClickEvent, TableEditEvent
+
 
 class BaseTable(ReactiveData, Widget):
 
@@ -74,17 +95,20 @@ class BaseTable(ReactiveData, Widget):
 
     value = param.Parameter(default=None)
 
-    _data_params = ['value']
+    _data_params: ClassVar[List[str]] = ['value']
 
-    _manual_params = ['formatters', 'editors', 'widths', 'titles', 'value', 'show_index']
+    _manual_params: ClassVar[List[str]] = [
+        'formatters', 'editors', 'widths', 'titles', 'value', 'show_index'
+    ]
 
-    _rename = {'disabled': 'editable', 'selection': None}
+    _rename: ClassVar[Mapping[str, str | None]] = {'disabled': 'editable', 'selection': None}
 
     __abstract = True
 
     def __init__(self, value=None, **params):
         self._renamed_cols = {}
         self._filters = []
+        self._index_mapping = {}
         super().__init__(value=value, **params)
 
     @param.depends('value', watch=True, on_init=True)
@@ -96,7 +120,7 @@ class BaseTable(ReactiveData, Widget):
             str(col) if str(col) != col else col: col for col in self._get_fields()
         }
 
-    def _validate(self, *events):
+    def _validate(self, *events: param.parameterized.Event):
         if self.value is None:
             return
         cols = self.value.columns
@@ -110,7 +134,7 @@ class BaseTable(ReactiveData, Widget):
             msg['editable'] = not msg.pop('editable') and len(self.indexes) <= 1
         return msg
 
-    def _get_fields(self):
+    def _get_fields(self) -> List[str]:
         indexes = self.indexes
         col_names = list(self.value.columns)
         if not self.hierarchical or len(indexes) == 1:
@@ -119,7 +143,7 @@ class BaseTable(ReactiveData, Widget):
             col_names = indexes[-1:] + col_names
         return col_names
 
-    def _get_columns(self):
+    def _get_columns(self) -> List[TableColumn]:
         if self.value is None:
             return []
 
@@ -128,7 +152,7 @@ class BaseTable(ReactiveData, Widget):
         df = self.value.reset_index() if len(indexes) > 1 else self.value
         return self._get_column_definitions(fields, df)
 
-    def _get_column_definitions(self, col_names, df):
+    def _get_column_definitions(self, col_names: List[str], df: DataFrameType) -> List[TableColumn]:
         import pandas as pd
         indexes = self.indexes
         columns = []
@@ -146,6 +170,8 @@ class BaseTable(ReactiveData, Widget):
 
             col_kwargs = {}
             kind = data.dtype.kind
+            editor: CellEditor
+            formatter: CellFormatter
             if kind == 'i':
                 formatter = NumberFormatter(text_align='right')
                 editor = IntEditor()
@@ -205,9 +231,16 @@ class BaseTable(ReactiveData, Widget):
         return columns
 
     @updating
-    def _update_cds(self, *events):
+    def _update_cds(self, *events: param.parameterized.Event):
         old_processed = self._processed
         self._processed, data = self._get_data()
+        if self._processed is not None and not (isinstance(self._processed, list) and not self._processed):
+            self._index_mapping = {
+                i: index
+                for i, index in enumerate(self._processed.index)
+            }
+        else:
+            self._index_mapping = {}
         # If there is a selection we have to compute new index
         if self.selection and old_processed is not None:
             indexes = list(self._processed.index)
@@ -225,9 +258,12 @@ class BaseTable(ReactiveData, Widget):
         for ref, (m, _) in self._models.items():
             self._apply_update(events, msg, m.source, ref)
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         source = ColumnDataSource(data=self._data)
-        source.selected.indices = self.selection
+        source.selected.indices = self.selection # type: ignore
         properties = self._get_properties(source)
         model = self._widget_type(**properties)
         if root is None:
@@ -237,14 +273,17 @@ class BaseTable(ReactiveData, Widget):
         self._models[root.ref['id']] = (model, parent)
         return model
 
-    def _update_columns(self, event, model):
+    def _update_columns(self, event: param.parameterized.Event, model: Model):
         if event.name == 'value' and [c.field for c in model.columns] == self._get_fields():
             # Skip column update if the data has changed but the columns
             # have not
             return
         model.columns = self._get_columns()
 
-    def _manual_update(self, events, model, doc, root, parent, comm):
+    def _manual_update(
+        self, events: Tuple[param.parameterized.Event, ...], model: Model, doc: Document,
+        root: Model, parent: Optional[Model], comm: Optional[Comm]
+    ) -> None:
         for event in events:
             if event.type == 'triggered' and self._updating:
                 continue
@@ -257,21 +296,48 @@ class BaseTable(ReactiveData, Widget):
             else:
                 self._update_columns(event, model)
 
-    def _sort_df(self, df):
+    def _sort_df(self, df: DataFrameType) -> DataFrameType:
         if not self.sorters:
             return df
         fields = [self._renamed_cols.get(s['field'], s['field']) for s in self.sorters]
         ascending = [s['dir'] == 'asc' for s in self.sorters]
-        rename = 'index' in fields and df.index.name is None
-        if rename:
-            df.index.name = 'index'
-        df_sorted = df.sort_values(fields, ascending=ascending)
+
+        # Temporarily add _index_ column because Tabulator uses internal _index
+        # as additional sorter to break ties
+        df['_index_'] = np.arange(len(df)).astype(str)
+        fields.append('_index_')
+        ascending.append(True)
+
+        # Handle sort on index column if show_index=True
+        if self.show_index:
+            rename = 'index' in fields and df.index.name is None
+            if rename:
+                df.index.name = 'index'
+        else:
+            rename = False
+
+        def tabulator_sorter(col):
+            # Tabulator JS defines its own sorting algorithm:
+            # - strings's case isn't taken into account
+            if col.dtype.kind not in 'SUO':
+                return col
+            try:
+                return col.str.lower()
+            except Exception:
+                return col
+
+        df_sorted = df.sort_values(fields, ascending=ascending, kind='mergesort',
+                                  key=tabulator_sorter)
+
+        # Revert temporary changes to DataFrames
         if rename:
             df.index.name = None
             df_sorted.index.name = None
+        df.drop(columns=['_index_'], inplace=True)
+        df_sorted.drop(columns=['_index_'], inplace=True)
         return df_sorted
 
-    def _filter_dataframe(self, df):
+    def _filter_dataframe(self, df: DataFrameType) -> DataFrameType:
         """
         Filter the DataFrame.
 
@@ -287,6 +353,8 @@ class BaseTable(ReactiveData, Widget):
         """
         filters = []
         for col_name, filt in self._filters:
+            if col_name is not None and col_name not in df.columns:
+                continue
             if isinstance(filt, (FunctionType, MethodType)):
                 df = filt(df)
                 continue
@@ -295,7 +363,9 @@ class BaseTable(ReactiveData, Widget):
             else:
                 val = filt
             column = df[col_name]
-            if np.isscalar(val):
+            if val is None:
+                continue
+            elif np.isscalar(val):
                 mask = column == val
             elif isinstance(val, (list, set)):
                 if not val:
@@ -317,6 +387,20 @@ class BaseTable(ReactiveData, Widget):
                                  "tuple or list.")
             filters.append(mask)
 
+        filters.extend(self._get_header_filters(df))
+
+        if filters:
+            mask = filters[0]
+            for f in filters:
+                mask &= f
+            if self._edited_indexes:
+                edited_mask = (df.index.isin(self._edited_indexes))
+                mask = mask | edited_mask
+            df = df[mask]
+        return df
+
+    def _get_header_filters(self, df):
+        filters = []
         for filt in getattr(self, 'filters', []):
             col_name = filt['field']
             op = filt['type']
@@ -329,6 +413,8 @@ class BaseTable(ReactiveData, Widget):
                     col = df.index
                 else:
                     col = df.index.get_level_values(self.indexes.index(col_name))
+            else:
+                continue
 
             # Sometimes Tabulator will provide a zero/single element list
             if isinstance(val, list):
@@ -374,12 +460,7 @@ class BaseTable(ReactiveData, Widget):
                 raise ValueError("Regex filtering not supported.")
             else:
                 raise ValueError(f"Filter type {op!r} not recognized.")
-        if filters:
-            mask = filters[0]
-            for f in filters:
-                mask &= f
-            df = df[mask]
-        return df
+        return filters
 
     def add_filter(self, filter, column=None):
         """
@@ -444,9 +525,12 @@ class BaseTable(ReactiveData, Widget):
             return [str(v) for v in values]
         return values
 
-    def _get_data(self):
+    def _get_data(self) -> Tuple[DataFrameType, DataDict]:
+        return self._process_df_and_convert_to_cds(self.value)
+
+    def _process_df_and_convert_to_cds(self, df: DataFrameType) -> Tuple[DataFrameType, DataDict]:
         import pandas as pd
-        df = self._filter_dataframe(self.value)
+        df = self._filter_dataframe(df)
         if df is None:
             return [], {}
         if isinstance(self.value.index, pd.MultiIndex):
@@ -564,10 +648,10 @@ class BaseTable(ReactiveData, Widget):
                 self.param.trigger('value')
             finally:
                 self._updating = False
-            stream_value = self._filter_dataframe(stream_value)
+            stream_value, stream_data = self._process_df_and_convert_to_cds(stream_value)
             try:
                 self._updating = True
-                self._stream(stream_value, rollover)
+                self._stream(stream_data, rollover)
             finally:
                 self._updating = False
         elif isinstance(stream_value, pd.Series):
@@ -575,10 +659,10 @@ class BaseTable(ReactiveData, Widget):
             if rollover is not None and len(self.value) > rollover:
                 with param.discard_events(self):
                     self.value = self.value.iloc[-rollover:]
-            stream_value = self._filter_dataframe(self.value.iloc[-1:])
+            stream_value, stream_data = self._process_df_and_convert_to_cds(self.value.iloc[-1:])
             try:
                 self._updating = True
-                self._stream(stream_value, rollover)
+                self._stream(stream_data, rollover)
             finally:
                 self._updating = False
         elif isinstance(stream_value, dict):
@@ -643,7 +727,7 @@ class BaseTable(ReactiveData, Widget):
         {'x': [3, 4], 'y': ['c', 'd']}
         """
         if self.value is None:
-            raise ValueError("Cannot patch empty {type(self).__name__}.")
+            raise ValueError(f"Cannot patch empty {type(self).__name__}.")
 
         import pandas as pd
         if not isinstance(self.value, pd.DataFrame):
@@ -776,20 +860,20 @@ class DataFrame(BaseTable):
         return to natural order. Use Shift + click to sort multiple
         columns simultaneously.""")
 
-    _manual_params = BaseTable._manual_params + ['aggregators']
+    _manual_params: ClassVar[List[str]] = BaseTable._manual_params + ['aggregators']
 
     _aggregators = {'sum': SumAggregator, 'max': MaxAggregator,
                     'min': MinAggregator, 'mean': AvgAggregator}
 
-    _source_transforms = {'hierarchical': None}
+    _source_transforms: ClassVar[Mapping[str, str | None]] = {'hierarchical': None}
 
-    _rename = {
+    _rename: ClassVar[Mapping[str, str | None]] = {
         'disabled': 'editable', 'selection': None, 'sorters': None,
         'text_align': None
     }
 
     @property
-    def _widget_type(self):
+    def _widget_type(self) -> Type[Model]:
         return DataCube if self.hierarchical else DataTable
 
     def _get_columns(self):
@@ -970,24 +1054,29 @@ class Tabulator(BaseTable):
         ], doc="""
         Tabulator CSS theme to apply to table.""")
 
-    _data_params = ['value', 'page', 'page_size', 'pagination', 'sorters', 'filters']
+    _data_params: ClassVar[List[str]] = [
+        'value', 'page', 'page_size', 'pagination', 'sorters', 'filters'
+    ]
 
-    _config_params = ['frozen_columns', 'groups', 'selectable', 'hierarchical']
+    _config_params: ClassVar[List[str]] = [
+        'frozen_columns', 'groups', 'selectable', 'hierarchical'
+    ]
 
-    _content_params = _data_params + ['expanded', 'row_content', 'embed_content']
+    _content_params: ClassVar[List[str]] = _data_params + ['expanded', 'row_content', 'embed_content']
 
-    _manual_params = BaseTable._manual_params + _config_params
+    _manual_params: ClassVar[List[str]] = BaseTable._manual_params + _config_params
 
-    _priority_changes = ['data']
+    _priority_changes: ClassVar[List[str]] = ['data']
 
-    _rename = {
+    _rename: ClassVar[Mapping[str, str | None]] = {
         'disabled': 'editable', 'selection': None, 'selectable': 'select_mode',
-        'row_content': None
+        'row_content': None, 'row_height': None, 'text_align': None,
+        'embed_content': None, 'header_align': None, 'header_filters': None
     }
 
     # Determines the maximum size limits beyond which (local, remote)
     # pagination is enabled
-    _MAX_ROW_LIMITS = (200, 10000)
+    _MAX_ROW_LIMITS: ClassVar[Tuple[int, int]] = (200, 10000)
 
     def __init__(self, value=None, **params):
         import pandas.io.formats.style
@@ -1000,9 +1089,11 @@ class Tabulator(BaseTable):
         self.style = None
         self._computed_styler = None
         self._child_panels = {}
+        self._edited_indexes = []
         self._explicit_pagination = 'pagination' in params
         self._on_edit_callbacks = []
         self._on_click_callbacks = {}
+        self._old_value = None
         super().__init__(value=value, **params)
         self._configuration = configuration
         self.param.watch(self._update_children, self._content_params)
@@ -1029,6 +1120,18 @@ class Tabulator(BaseTable):
     def _set_explicict_pagination(self):
         self._explicit_pagination = True
 
+    @staticmethod
+    def _validate_iloc(idx, iloc):
+        # Validate that the index returned by Pandas get_loc is a single int,
+        # as get_loc can return a slice or a mask array when it finds more
+        # than one locations.
+        if not isinstance(iloc, int):
+            raise ValueError(
+                'The Tabulator widget expects the provided `value` Pandas DataFrame '
+                'to have unique indexes, in particular when it has to deal with '
+                f'click or edit events. Found this duplicate index: {idx!r}'
+            )
+
     def _validate(self, *events):
         super()._validate(*events)
         if self.value is not None:
@@ -1041,36 +1144,52 @@ class Tabulator(BaseTable):
             except Exception:
                 pass
 
-    def _cleanup(self, root):
+    def _cleanup(self, root: Model | None = None) -> None:
         for p in self._child_panels.values():
             p._cleanup(root)
         super()._cleanup(root)
 
     def _process_event(self, event):
         event_col = self._renamed_cols.get(event.column, event.column)
-        if self.pagination == 'remote':
+        if self.pagination in ['remote']:
             nrows = self.page_size
             event.row = event.row+(self.page-1)*nrows
+
+        idx = self._index_mapping[event.row]
+        iloc = self.value.index.get_loc(idx)
+        self._validate_iloc(idx, iloc)
+        event.row = iloc
         if event_col not in self.buttons:
-            if event_col not in self._processed.columns:
-                event.value = self._processed.index[event.row]
+            if event_col not in self.value.columns:
+                event.value = self.value.index[event.row]
             else:
-                event.value = self._processed[event_col].iloc[event.row]
+                event.value = self.value[event_col].iloc[event.row]
+
+        # Check if edited cell was filtered
+        import pandas as pd
+        filter_df = pd.DataFrame([event.value], columns=[event.column])
+        filters = self._get_header_filters(filter_df)
+        if filters and filters[0].any():
+            self._edited_indexes.append(idx)
+
+        # Set the old attribute on a table edit event
+        if event.event_name == 'table-edit' and self._old_value is not None:
+            event.old = self._old_value[event_col].iloc[event.row]
         if event.event_name == 'table-edit':
-            if self._old is not None:
-                event.old = self._old[event_col].iloc[event.row]
             for cb in self._on_edit_callbacks:
-                state.execute(partial(cb, event))
+                state.execute(partial(cb, event), schedule=False)
             self._update_style()
         else:
             for cb in self._on_click_callbacks.get(None, []):
-                state.execute(partial(cb, event))
+                state.execute(partial(cb, event), schedule=False)
             for cb in self._on_click_callbacks.get(event_col, []):
-                state.execute(partial(cb, event))
+                state.execute(partial(cb, event), schedule=False)
 
     def _get_theme(self, theme, resources=None):
         from ..io.resources import RESOURCE_MODE
-        from ..models.tabulator import _get_theme_url, THEME_PATH, THEME_URL
+        from ..models.tabulator import (
+            _TABULATOR_THEMES_MAPPING, THEME_PATH, THEME_URL, _get_theme_url,
+        )
         if RESOURCE_MODE == 'server' and resources in (None, 'server'):
             theme_url = f'{LOCAL_DIST}bundled/datatabulator/{THEME_PATH}'
             if state.rel_path:
@@ -1080,7 +1199,8 @@ class Tabulator(BaseTable):
         # Ensure theme_url updates before theme
         cdn_url = _get_theme_url(THEME_URL, theme)
         theme_url = _get_theme_url(theme_url, theme)
-        fname = 'tabulator' if self.theme == 'default' else 'tabulator_'+self.theme
+        theme_ = _TABULATOR_THEMES_MAPPING.get(self.theme, self.theme)
+        fname = 'tabulator' if theme_ == 'default' else 'tabulator_' + theme_
         if self._widget_type is not None:
             self._widget_type.__css_raw__ = [f'{cdn_url}{fname}.min.css']
         return theme_url, theme
@@ -1108,6 +1228,35 @@ class Tabulator(BaseTable):
                 return
         model.configuration = self._get_configuration(model.columns)
 
+    def _process_data(self, data):
+        # Extending _process_data to cover the case when header filters are
+        # active and a cell is edited. In that case the data received from the
+        # front-end is the full table, not just the filtered one. However the
+        # _processed data is already filtered, this made the comparison between
+        # the new data and old data wrong. This extension replicates the
+        # front-end filtering - if need be - to be able to correctly make the
+        # comparison and update the data held by the backend.
+
+        # It also makes a copy of the value dataframe, to use it to obtain
+        # the old value in a table-edit event.
+        self._old_value = self.value.copy()
+
+        import pandas as pd
+        df = pd.DataFrame(data)
+        filters = self._get_header_filters(df)
+        if filters:
+            mask = filters[0]
+            for f in filters:
+                mask &= f
+            if self._edited_indexes:
+                edited_mask = (df['index'].isin(self._edited_indexes))
+                mask = mask | edited_mask
+            df = df[mask]
+        data = {
+            col: df[col].values for col in df.columns
+        }
+        return super()._process_data(data)
+
     def _get_data(self):
         if self.pagination != 'remote' or self.value is None:
             return super()._get_data()
@@ -1116,6 +1265,7 @@ class Tabulator(BaseTable):
         df = self._sort_df(df)
         nrows = self.page_size
         start = (self.page-1)*nrows
+
         page_df = df.iloc[start: start+nrows]
         if isinstance(self.value.index, pd.MultiIndex):
             indexes = [
@@ -1243,7 +1393,7 @@ class Tabulator(BaseTable):
         if self.pagination == 'remote':
             length = self._length
             nrows = self.page_size
-            max_page = length//nrows + bool(length%nrows)
+            max_page = max(length//nrows + bool(length%nrows), 1)
             if self.page != max_page:
                 return
         super()._stream(stream, rollover)
@@ -1256,8 +1406,10 @@ class Tabulator(BaseTable):
         if follow and self.pagination:
             length = self._length
             nrows = self.page_size
-            self.page = length//nrows + bool(length%nrows)
+            self.page = max(length//nrows + bool(length%nrows), 1)
         super().stream(stream_value, rollover, reset_index)
+        if follow and self.pagination:
+            self._update_max_page()
 
     @updating
     def _patch(self, patch):
@@ -1283,10 +1435,14 @@ class Tabulator(BaseTable):
         self._update_selectable()
 
     def _update_cds(self, *events):
+        if any(event.name == 'filters' for event in events):
+            self._edited_indexes = []
         page_events = ('page', 'page_size', 'sorters', 'filters')
         if self._updating:
             return
-        elif (events and all(e.name in page_events for e in events) and not self.pagination):
+        elif events and all(e.name in page_events[:-1] for e in events) and self.pagination == 'local':
+            return
+        elif events and all(e.name in page_events for e in events) and not self.pagination:
             self._processed, _ = self._get_data()
             return
         recompute = not all(
@@ -1307,19 +1463,21 @@ class Tabulator(BaseTable):
     def _update_max_page(self):
         length = self._length
         nrows = self.page_size
-        max_page = length//nrows + bool(length%nrows)
+        max_page = max(length//nrows + bool(length%nrows), 1)
         self.param.page.bounds = (1, max_page)
         for ref, (model, _) in self._models.items():
             self._apply_update([], {'max_page': max_page}, model, ref)
 
-    def _update_selected(self, *events, indices=None):
+    def _update_selected(self, *events: param.parameterized.Event, indices=None):
         kwargs = {}
         if self.pagination == 'remote' and self.value is not None:
             index = self.value.iloc[self.selection].index
             indices = []
             for v in index.values:
                 try:
-                    indices.append(self._processed.index.get_loc(v))
+                    iloc = self._processed.index.get_loc(v)
+                    self._validate_iloc(v ,iloc)
+                    indices.append(iloc)
                 except KeyError:
                     continue
             nrows = self.page_size
@@ -1329,7 +1487,7 @@ class Tabulator(BaseTable):
                                  if ind>=start and ind<end]
         super()._update_selected(*events, **kwargs)
 
-    def _update_column(self, column, array):
+    def _update_column(self, column: str, array: np.ndarray):
         if self.pagination != 'remote':
             index = self._processed.index.values
             self.value.loc[index, column] = array
@@ -1342,7 +1500,7 @@ class Tabulator(BaseTable):
         self.value[column].loc[index] = array
         self._processed[column].loc[index] = array
 
-    def _update_selection(self, indices):
+    def _update_selection(self, indices: List[int]):
         if self.pagination != 'remote':
             self.selection = indices
             return
@@ -1352,12 +1510,14 @@ class Tabulator(BaseTable):
         indices = []
         for v in index.values:
             try:
-                indices.append(self.value.index.get_loc(v))
+                iloc = self.value.index.get_loc(v)
+                self._validate_iloc(v, iloc)
+                indices.append(iloc)
             except KeyError:
                 continue
         self.selection = indices
 
-    def _get_properties(self, source):
+    def _get_properties(self, source: ColumnDataSource) -> Dict[str, Any]:
         props = {p : getattr(self, p) for p in list(Layoutable.param)
                  if getattr(self, p) is not None}
         columns = self._get_columns()
@@ -1388,14 +1548,17 @@ class Tabulator(BaseTable):
         props.update(self._process_param_change(process))
         if self.pagination:
             length = 0 if self._processed is None else len(self._processed)
-            props['max_page'] = length//self.page_size + bool(length%self.page_size)
+            props['max_page'] = max(length//self.page_size + bool(length%self.page_size), 1)
         if self.frozen_rows and 'height' not in props and 'sizing_mode' not in props:
             length = self.page_size+2 if self.pagination else self._length
             props['height'] = min([length * self.row_height + 30, 2000])
         props['indexes'] = self.indexes
         return props
 
-    def _get_model(self, doc, root=None, parent=None, comm=None):
+    def _get_model(
+        self, doc: Document, root: Optional[Model] = None,
+        parent: Optional[Model] = None, comm: Optional[Comm] = None
+    ) -> Model:
         if self._widget_type is None:
             self._widget_type = lazy_load(
                 'panel.models.tabulator', 'DataTabulator', isinstance(comm, JupyterComm), root
@@ -1412,22 +1575,20 @@ class Tabulator(BaseTable):
             child_panels, doc, root, parent, comm
         )
         self._link_props(model, ['page', 'sorters', 'expanded', 'filters'], doc, root, comm)
-        if comm:
-            model.on_event('table-edit', self._comm_event)
-            model.on_event('cell-click', self._comm_event)
-        else:
-            model.on_event('table-edit', partial(self._server_event, doc))
-            model.on_event('cell-click', partial(self._server_event, doc))
+        self._register_events('cell-click', 'table-edit', model=model, doc=doc, comm=comm)
         return model
 
-    def _update_model(self, events, msg, root, model, doc, comm):
+    def _update_model(
+        self, events: Dict[str, param.parameterized.Event], msg: Dict[str, Any],
+        root: Model, model: Model, doc: Document, comm: Optional[Comm]
+    ) -> None:
         if comm and 'theme_url' in msg:
             msg['theme_url'], msg['theme'] = self._get_theme(
                 msg.pop('theme'), 'inline'
             )
         super()._update_model(events, msg, root, model, doc, comm)
 
-    def _get_filter_spec(self, column):
+    def _get_filter_spec(self, column: TableColumn) -> Dict[str, Any]:
         fspec = {}
         if not self.header_filters or (isinstance(self.header_filters, dict) and
                                        column.field not in self.header_filters):
@@ -1443,8 +1604,15 @@ class Tabulator(BaseTable):
                 elif col.dtype.kind == 'b':
                     fspec['headerFilter'] = 'tickCross'
                     fspec['headerFilterParams'] = {'tristate': True, 'indeterminateValue': None}
+                elif isdatetime(col) or col.dtype.kind == 'M':
+                    fspec['headerFilter'] = False
                 else:
                     fspec['headerFilter'] = True
+            elif isinstance(column.editor, DateEditor):
+                # Datetime filtering currently broken with Tabulator 5.4.3
+                # Initial (empty) value of filter is passed to luxon.js
+                # and causes error
+                fspec['headerFilter'] = False
             else:
                 fspec['headerFilter'] = True
             return fspec
@@ -1458,9 +1626,24 @@ class Tabulator(BaseTable):
             filter_params = {}
             filter_func = None
             filter_placeholder = None
-        fspec['headerFilter'] = filter_type
-        if filter_type == 'select' and not filter_params:
-            filter_params = {'values': True}
+        # Tabulator JS renamed select and autocomplete to list, and relies on
+        # valuesLookup set to True to autopopulate the filter, instead of
+        # values. This ensure backwards compatibility.
+        if filter_type in ['select', 'autocomplete']:
+            self.param.warning(
+                f'The {filter_type!r} filter has been deprecated, use '
+                f'instead the "list" filter type to configure column {column.field!r}'
+            )
+            filter_type = 'list'
+            if filter_params.get('values', False) is True:
+                self.param.warning(
+                    'Setting "values" to True has been deprecated, instead '
+                    f'set "valuesLookup" to True to configure column {column.field!r}'
+                )
+                del filter_params['values']
+                filter_params['valuesLookup'] = True
+        if filter_type == 'list' and not filter_params:
+            filter_params = {'valuesLookup': True}
         fspec['headerFilter'] = filter_type
         if filter_params:
             fspec['headerFilterParams'] = filter_params
@@ -1470,7 +1653,7 @@ class Tabulator(BaseTable):
             fspec['headerFilterPlaceholder'] = filter_placeholder
         return fspec
 
-    def _config_columns(self, column_objs):
+    def _config_columns(self, column_objs: List[TableColumn]) -> List[Dict[str, Any]]:
         column_objs = list(column_objs)
         groups = {}
         columns = []
@@ -1526,6 +1709,16 @@ class Tabulator(BaseTable):
                 formatter = dict(formatter)
                 col_dict['formatter'] = formatter.pop('type')
                 col_dict['formatterParams'] = formatter
+            col_name = self._renamed_cols[column.field]
+            if column.field in self.indexes:
+                if len(self.indexes) == 1:
+                    dtype = self.value.index.dtype
+                else:
+                    dtype = self.value.index.get_level_values(self.indexes.index(column.field)).dtype
+            else:
+                dtype = self.value.dtypes[col_name]
+            if dtype.kind == 'M':
+                col_dict['sorter'] = 'timestamp'
             editor = self.editors.get(column.field)
             if column.field in self.editors and editor is None:
                 col_dict['editable'] = False
@@ -1535,6 +1728,15 @@ class Tabulator(BaseTable):
                 editor = dict(editor)
                 col_dict['editor'] = editor.pop('type')
                 col_dict['editorParams'] = editor
+            if col_dict.get('editor') in ['select', 'autocomplete']:
+                self.param.warning(
+                    f'The {col_dict["editor"]!r} editor has been deprecated, use '
+                    f'instead the "list" editor type to configure column {column.field!r}'
+                )
+                col_dict['editor'] = 'list'
+                if col_dict.get('editorParams', {}).get('values', False) is True:
+                    del col_dict['editorParams']['values']
+                    col_dict['editorParams']['valuesLookup']
             if column.field in self.frozen_columns or i in self.frozen_columns:
                 col_dict['frozen'] = True
             if isinstance(self.widths, dict) and isinstance(self.widths.get(column.field), str):
@@ -1555,7 +1757,7 @@ class Tabulator(BaseTable):
                 columns.append(col_dict)
         return columns
 
-    def _get_configuration(self, columns):
+    def _get_configuration(self, columns: List[Dict[str, Any]]) -> Dict[str, Any]:
         """
         Returns the Tabulator configuration.
         """
@@ -1573,7 +1775,7 @@ class Tabulator(BaseTable):
             configuration['height'] = self.height
         return configuration
 
-    def download(self, filename='table.csv'):
+    def download(self, filename: str = 'table.csv'):
         """
         Triggers downloading of the table as a CSV or JSON.
 
@@ -1583,8 +1785,8 @@ class Tabulator(BaseTable):
             The filename to save the table as.
         """
         for ref, (model, _) in self._models.items():
-            self._apply_update([], {'filename': filename}, model, ref)
-            self._apply_update([], {'download': not model.download}, model, ref)
+            self._apply_update({}, {'filename': filename}, model, ref)
+            self._apply_update({}, {'download': not model.download}, model, ref)
 
     def download_menu(self, text_kwargs={}, button_kwargs={}):
         """
@@ -1622,7 +1824,7 @@ class Tabulator(BaseTable):
         """)
         return filename, button
 
-    def on_edit(self, callback):
+    def on_edit(self, callback: Callable[[TableEditEvent], None]):
         """
         Register a callback to be executed when a cell is edited.
         Whenever a cell is edited on_edit callbacks are called with
@@ -1636,7 +1838,7 @@ class Tabulator(BaseTable):
         """
         self._on_edit_callbacks.append(callback)
 
-    def on_click(self, callback, column=None):
+    def on_click(self, callback: Callable[[CellClickEvent], None], column: Optional[str] = None):
         """
         Register a callback to be executed when any cell is clicked.
         The callback is given a CellClickEvent declaring the column
@@ -1655,7 +1857,7 @@ class Tabulator(BaseTable):
         self._on_click_callbacks[column].append(callback)
 
     @property
-    def current_view(self):
+    def current_view(self) -> DataFrameType:
         """
         Returns the current view of the table after filtering and
         sorting are applied.
